#----------------------------------------------------------------------
# AbsenceDB - interface to database
# $Id: AbsenceDB.pm 91 2009-07-14 15:58:06Z urban $
#----------------------------------------------------------------------

# $Rev: 91 $:  Revision of last commit
# $Author: urban $:  Author of last commit
# $Date: 2009-07-14 17:58:06 +0200 (Tue, 14 Jul 2009) $:  Date of last commit

package AbsenceDB;

use FileHandle;
use Carp;
use Digest::MD5 qw(md5_base64);

use AbsenceConfig;
use AbsenceDate;
use AbsenceLog;
use DBI;
use CGI::Session;
use IPC::Open3;

#----------------------------------------------------------------------
# initialization
#----------------------------------------------------------------------

my $DB_NAME			= AbsenceConfig::fetch('database_name');
my $DB_PORT			= AbsenceConfig::fetch('database_port');
my $DB_HOST			= AbsenceConfig::fetch('database_host');
my $DB_USER			= AbsenceConfig::fetch('database_user');
my $DB_PASS			= AbsenceConfig::fetch('database_pass');

my $DEBUG			= 0;
my $VERBOSE			= 1;
my %META;	# container for PostgreSQL metadata about tables/fields

#----------------------------------------------------------------------
# connect to db first thing
#----------------------------------------------------------------------
$DEBUG && abslog("init: connect to DB");

my $DBH = DBI->connect("dbi:Pg:dbname=$DB_NAME;host=$DB_HOST",
	$DB_USER, $DB_PASS, {AutoCommit => 1, RaiseError => 1});

defined($DBH) || die "cannot connect to DB";

my @SESSION_PARAMS	= (
	'driver:PostgreSQL',
	undef,
	{
		Handle		=> $DBH,
		ColumnType	=> 'binary',
	}
);

END {
	defined($DBH) && $DBH->disconnect;
}

#-----------------------------------------
# $FULL_SLOTS is defined here *and* in
# AbsenceImage.pm :(
#-----------------------------------------
my $FULL_SLOTS		= 4;

my $MODTIME_FILE	= AbsenceConfig::fetch('modtime_db_file');
my $OAP				= AbsenceConfig::fetch('objects_are_people');
my $AUTH			= AbsenceConfig::fetch('authentication');
my $CRED_SRC		= AbsenceConfig::fetch('credential_src');
my $PW_HASH_FORMAT	= AbsenceConfig::fetch('pw_hash_format');
my $SESSION_TIMEOUT	= AbsenceConfig::fetch('session_timeout');
my $MULTI_RES		= AbsenceConfig::fetch('multi_res');
my $MAX_MULTI		= AbsenceConfig::fetch('max_multi');


my $VERSION			= '1.5';

#----------------------------------------------------------------------
# methods
#----------------------------------------------------------------------

#----------------------------------------------------------------
# getPeople()
#
# returns an array of person_ids sorted by name
# if group_id is zero, returns all people. Otherwise, returns
# people in group_id
#----------------------------------------------------------------

sub getPeople
{
	my $group_id = shift;

	my @out;
	my $sql;

	if ($group_id == 0) {
		$sql = qq{
			SELECT id
			FROM v_object
			ORDER BY LOWER(name);
		};
		return dbSelectColumn($sql, 'id');
	}

	if (!groupExists(id => $group_id)) {
		die "getPeople: gid [$group_id] does not exist";
	}

	$sql = qq{
		SELECT o.id
		FROM
			v_object AS o,
			v_group AS g,
			c_group_object AS go
		WHERE
			g.id = ? AND
			go.group_id = g.id AND
			go.object_id = o.id
		ORDER BY
			LOWER(o.name);
	};
	return dbSelectColumn($sql, 'id', $group_id);
}

sub groupExists
{
	my ($field, $val) = @_;

	my $sql = qq{SELECT id FROM v_group WHERE $field = ?;};

	my $rv = $DBH->do($sql, undef, $val);
	($rv eq '0E0') && return 0;
	return 1;
}

sub reservationExists
{
	my $id = shift;

	my $sql = qq{SELECT id FROM v_reservation WHERE id = ?;};

	my $rv = $DBH->do($sql, undef, $id);
	($rv eq '0E0') && return 0;
	return 1;
}

sub objectExists
{
	my ($field, $val) = @_;

	my $sql = qq{SELECT id FROM v_object WHERE $field = ?;};

	my $rv = $DBH->do($sql, undef, $val);
	($rv eq '0E0') && return 0;
	return 1;
}

sub userExists
{
	my ($field, $val) = @_;

	my $sql = qq{SELECT id FROM v_user WHERE $field = ?;};

	my $rv = $DBH->do($sql, undef, $val);
	($rv eq '0E0') && return 0;
	return 1;
}

sub inListN
{
	my ($thing, $lref) = @_;

	foreach my $elem (@{$lref}) {
		($elem == $thing) && return 1;
	}

	return 0;
}

sub getPerson
{
	my ($pid, $field) = @_;

	if (!objectExists(id => $pid)) {
		#print "no person with ID=[$pid]\n";
		return undef;
	}
	
	my $sql;

	if ($field) {
		$sql = qq{SELECT $field FROM v_object WHERE id = ?;};

		my $ref = dbSelect($sql, $pid);
		return defined($ref) ? $ref->{$field} : undef;
	}

	$sql = qq{SELECT * FROM v_object WHERE id = ?;};

	return dbSelect($sql, $pid);
}

sub getUserForOid
{
	my $oid = shift;

	my $sql = qq{
		SELECT *
		FROM v_user
		WHERE object_id = ?;
	};

	return dbSelect($sql, $oid);
}

sub getOidForUid
{
	my $user_id = shift;

	my $sql = qq{
		SELECT	object_id
		FROM	v_user
		WHERE	id = ?;
	};

	my $res = dbSelect($sql, $user_id);

	return defined($res) ? $res->{object_id} : undef;
}

sub getPidForUser
{
	my $user = shift;

	userExists(username => $user) || confess "user [$user] unknown";

	my $sql = qq{
		SELECT	object_id
		FROM	v_user
		WHERE	username = ?;
	};

	my $res = dbSelect($sql, $user);

	return defined($res) ? $res->{object_id} : undef;
}

#-------------------------------------------------------------
# getReservations()
#
# apparently only called from resConflict, probably don't need anymore
# also, inefficient, as it queries all reservations, without regard
# for date
#-------------------------------------------------------------
sub getReservations
{
	my $person_id = shift;

	my @out;

	$DEBUG && abslog("getReservations for id=[$person_id]");

	my $sql = qq{
		SELECT *
		FROM v_reservation
		WHERE object_id = ?;
	};

	return dbSelect($sql, $person_id);
}

sub getMonthReservations
{
	my ($person_id, $month, $year) = @_;

	my $dim = AbsenceDate::daysInMonth($month, $year);
	my $month_begin = "$year-$month-1";
	my $month_end = "$year-$month-$dim";

	#$DEBUG && abslog("getMonthReservations: p_id = $person_id, m=$month, y=$year");

	my $sql = qq{
		SELECT *
		FROM v_reservation
		WHERE
			object_id = ? AND (
				(
					( start BETWEEN ? AND ? ) OR
					( finish BETWEEN ? AND ? )
				)
				OR
				(
					( start < ? ) AND ( finish > ? )
				)
			);
	};

	my @bind_params = (
		$person_id,
		$month_begin, $month_end, $month_begin, $month_end,
		$month_begin, $month_end);

	my @out;
	foreach my $ref (dbSelect($sql, @bind_params)) {
		my ($sd, $ed) = getMonthBounds($ref->{start}, $ref->{finish}, $month, $year);
		push(@out, {
			res		=> $ref,
			bounds	=> {
				start	=> $sd,
				end		=> $ed,
			 },
		});
	}

	@out;
}

sub getMonthBounds
{
	my ($start, $end, $month, $year) = @_;

	my ($sd, $sm, $sy, $ed, $em, $ey);

	if (ref($start)) {
		($sy, $sm, $sd) = ($start->{year}, $start->{month}, $start->{day});
		($ey, $em, $ed) = ($end->{year}, $end->{month}, $end->{day});
	} else {
		($sy, $sm, $sd) = ($start =~ /^(\d{4})-(\d{2})-(\d{2})$/);
		($ey, $em, $ed) = ($end =~ /^(\d{4})-(\d{2})-(\d{2})$/);
	}

	my $dim = AbsenceDate::daysInMonth($month, $year);

	my $s = (($sm == $month) && ($sy == $year)) ? $sd : 1;
	my $e = (($em == $month) && ($ey == $year)) ? $ed : $dim;

	return ($s, $e);
}

sub getReservation
{
	my $rid = shift;

	my $sql = qq{
		SELECT *
		FROM v_reservation
		WHERE id = ?;
	};

	return dbSelect($sql, $rid);
}

#--------------------------------------------------------------------
# group stuff
#--------------------------------------------------------------------

sub addGroup
{
	# ROBBO: modify callers to pass desc too
	my ($gid, $name, $pass, $desc) = @_;

	$DEBUG && abslog("addGroup: start: name=[$name], pass=[$pass]");

	## check lengths
	#my $ret = checkLengths('c_group', {
	#	name		=> $name,
	#	password	=> $pass,
	#	description	=> $desc,
	#};
	#$ret && return "baddata: $ret";

	my $verb;

	my $sql;

	if ($gid eq 'new') {
		if (groupExists(name => $name)) {
			abslog("addGroup: dup, bailing.");
			return 'duplicate';
		}
		$sql = qq{
			INSERT INTO c_group
				(name, password, description)
			VALUES (?,?,?);
		};
		$verb = 'insert';
		$gid = insertRowRetrieveSequence('c_group', $sql, $name, $pass, $desc);

	} else {
		if (!groupExists(id => $gid)) {
			abslog("error: attempt to modify non-existent group [$gid]");
			return undef;
		}
		$sql = qq{
			UPDATE c_group
			SET name = ?, password = ?, description = ?
			WHERE id = ?;
		};
		$verb = 'update';
		my $rv = $DBH->do($sql, undef, $name, $pass, $desc, $gid);
		defined($rv) || die "update group [$name]. error=".$DBH->errstr();
	}

	$DEBUG && abslog("addGroup: group-id = [$gid].");
	abslog("addGroup: $verb: name=[$name], gid=[$gid]");

	return wantarray ? ('ok', $gid) : 'ok';
}

sub deleteGroup
{
	my $gid = shift;

	$DEBUG && abslog("deleteGroup: start: gid=[$gid]");

	my $sql = qq{SELECT * FROM v_group WHERE id = $gid;};
	my $gref = dbSelect($sql);

	if (!defined($gref)) {
		abslog("error: attempt to delete non-existent group [$gid]");
		return undef;
	}

	#---------------------------------------------------
	# first, check if members need to be deleted or modified
	#---------------------------------------------------
	$sql = qq{
		SELECT DISTINCT v_object.id
		FROM v_group,v_object,c_group_object
		WHERE c_group_object.group_id = ?
		AND c_group_object.object_id = v_object.id;
	};

	my $aref = $DBH->selectcol_arrayref($sql, undef, $gid);
	if (defined($aref)) {
		#---------------------------------------------------
		# people are still members of $gid
		# for each member, check if member of other groups
		# if yes, delete membership, if no, delete object
		#---------------------------------------------------
		foreach my $pid (@{$aref}) {
			# get IDs of object/group mapping for $pid
			removeObjectFromGroup(
				-pid	=> $pid,
				-gid	=> $gid,
				#'-delete_groupless_objects',
			);
		}
	}

	#---------------------------------------------------
	# remove any referring rows in d_group_mod_time
	#---------------------------------------------------
	$sql = qq{ DELETE FROM d_group_mod_time WHERE group_id = ?; };
	$DBH->do($sql, undef, $gid);

	#-----------------------------------------------------------------
	# now get rid of the group
	# it might be interesting some time to implement invalidation.
	# for now, I just delete the row entirely.
	#-----------------------------------------------------------------
	#my $now = time();
	#$sql = qq{UPDATE c_group SET invalidated = int2tsz($now) WHERE id = $gid;};
	#-----------------------------------------------------------------

	$sql = qq{DELETE FROM c_group WHERE id = ?;};
	$DBH->do($sql, undef, $gid);
	
	$AUTH && deleteGidInAcls($gid);

	abslog("deleteGroup: name=[$gref->{name}], gid=[$gid]");

	return 'ok';
}

sub addGroupMappings
{
	my ($object_id, @gids) = @_;

	my $sql = qq{
		INSERT INTO c_group_object
			(group_id,object_id)
		VALUES (?,?);
	};
	my $sth = $DBH->prepare($sql);
	defined($sth) || die "prepare failed on sql [$sql]";

	#	VALUES ($gid,$object_id);
	foreach my $gid (@gids) {
		$sth->execute($gid, $object_id)
			|| die "sql failed: [$sql] with params gid=[$gid], object_id=[$object_id]";
	}
}

sub deleteGroupMappings
{
	my ($object_id, @group_ids) = @_;

	my $sql;
	my @bind_params;

	if (!@group_ids) {
		$sql = qq{
			DELETE FROM c_group_object
			WHERE object_id = ?;
		};
		@bind_params = ($object_id);
	} else {
		my @questions = map '?', @group_ids;
		my $q = join(',', @questions);
		$sql = qq{
			DELETE FROM c_group_object
			WHERE object_id = ?
			AND group_id in ($q);
		};
		@bind_params = ($object_id, @group_ids);
	}
	$DBH->do($sql, undef, @bind_params);
}

#--------------------------------------------------------------------
# removeObjectFromGroup()
#
# remove an object from a group by deleting the object-group mapping.
# parameters:
# -gid => $gid
# -pid => $pid
# -delete_groupless_objects
#
# $gid and $pid are mandatory
#
# returns: nothing
#--------------------------------------------------------------------
sub removeObjectFromGroup
{
	my $delete_groupless_objects = 0;
	my ($pid, $gid);

	while(my $arg = shift) {
		if ($arg =~ /delete_groupless_objects/) {
			$delete_groupless_objects = 1
		} elsif ($arg eq '-pid') {
			$pid = shift;
		} elsif ($arg eq '-gid') {
			$gid = shift;
		}
	}
	defined($pid) && defined($gid) || die "removeObjectFromGroup: pid/gid not defined";

	my $sql = qq{
		SELECT *
		FROM c_group_object
		WHERE c_group_object.object_id = ?;
	};
	my @list = dbSelect($sql, $pid);

	my $mapping_to_delete;
	my @others;
	foreach my $mapping (@list) {
		if ($mapping->{group_id} == $gid) {
			$mapping_to_delete = $mapping;
		} else {
			push(@others, $mapping);
		}
	}

	# $mapping_to_delete now contains id of the mapping which
	# associates object $pid with group $gid.
	# @others possibly contains mappings for $pid to other groups

	#--------------------------------------------------------
	# delete mapping
	#--------------------------------------------------------
	if (defined($mapping_to_delete)) {
		$sql = qq[
			DELETE FROM c_group_object
			WHERE id = ?;
		];
		$DBH->do($sql, undef, $mapping_to_delete->{id});
	}

	#--------------------------------------------------------
	# invalidate object, if desired, and if object is not member
	# of other groups
	#--------------------------------------------------------
	if ($delete_groupless_objects && !@others) {
		#invalidateObject($pid);
		_deleteObject($pid);
	}
}

sub bad_deleteObject
{
	my $oid = shift;
	
	deleteReservations(object_id => $oid);

	my $sql = qq{ DELETE FROM c_object WHERE id = ?; };
	$DBH->do($sql, undef, $oid);
}

sub invalidateObject
{
	my $oid = shift;
	
	invalidateReservations(object_id => $oid);

	my $now = time();

	my $sql = qq{
		UPDATE c_object SET invalidated = int2tsz($now) WHERE id = ?;
	};
	$DBH->do($sql, undef, $oid);
}

sub _deleteReservation
{
	my $res_id = shift;

	my $sql = qq{
		DELETE FROM d_reservation WHERE id = ?;
	};
	$DBH->do($sql, undef, $res_id);
}

sub invalidateReservation
{
	my $res_id = shift;

	my $now = time();

	my $sql = qq{
		UPDATE d_reservation SET invalidated = int2tsz($now) WHERE id = ?;
	};
	$DBH->do($sql, undef, $res_id);
}

sub deleteReservations
{
	my ($field, $val) = @_;

	my $sql = qq{
		DELETE FROM d_reservation WHERE $field = ?;
	};
	$DEBUG && abslog("deleteReservations: sql=[$sql]");

	$DBH->do($sql, undef, $val);
}

sub invalidateReservations
{
	my ($field, $val) = @_;

	my $now = time();

	my $sql;

	$sql = qq{
		UPDATE d_reservation SET invalidated = int2tsz($now) WHERE $field = ?;
	};

	$DBH->do($sql, undef, $val);
}

#--------------------------------------------------------------------
# ACL stuff
#--------------------------------------------------------------------

sub convertAclToInternal
{
	my @out;
	foreach (@_) {
		push(@out, {
			level	=> $_->{level},
			target	=> defined($_->{magic}) ? $_->{magic} : $_->{ref_id},
		});
	}
	return wantarray ? (@out) : $out[0];
}

sub getUserAcls
{
	my $user_id = shift;

	my $sql = qq{
		SELECT *
		FROM c_access
		WHERE user_id = ?;
	};

	my @acls = dbSelect($sql, $user_id);
	@acls || return undef;

	# separate ACLs into group/object

	my (@group, @object);

	foreach my $acl (@acls) {
		if ($acl->{type} eq 'object') {
			push(@object, convertAclToInternal($acl));
		} else {
			push(@group, convertAclToInternal($acl));
		}
	}

	return { object => \@object, group => \@group };
}

#------------------------------------------------------
# authorizeAccessToObject()
#
# determines whether a user, ($user_id) may write an object ($oid),
# that is, create, modify, and delete reservations.  a user
# may gain write privileges in one of four:
# (in all access entries level must be >= 2 and user_id must equal $user_id)
# 1. there is an object access entry with magic=self and pointing to
#    the oid associated with the user_id
# 2. there is an object access entry pointing to the oid
# 3. there is a group access entry pointing to a group to which
#    oid belongs
# 4. there is a group access entry with magic == 'all'
# 5. there is a group access entry with magic == self, and the
#    object associated with $user_id belongs to a group to which
#    the object $oid also belongs
#------------------------------------------------------
sub authorizeAccessToObject
{
	my ($user_id, $oid, $access) = @_;

	my %access_map = (
		read	=> 1,
		write	=> 2,
		admin	=> 4,
	);

	exists($access_map{$access}) || die "access-type [$access] uknown";

	my $sql = qq{
		SELECT *
		FROM
			c_access		AS a,
			c_group_object	AS map_u,
			c_group_object	AS map_o,
			v_user			AS u
		WHERE
			(a.user_id = ? AND a.level >= $access_map{$access}) AND
			(
					(a.type				= 'object'			AND
					 a.magic			= 'self'			AND
					 u.id				= ?					AND
					 u.object_id		= ?) 
				OR
					(a.type				= 'object'			AND
					 a.ref_id			= ?)
				OR
					(a.type				= 'group'			AND 
					 a.ref_id			= map_o.group_id	AND
					 map_o.object_id	= ?)
				OR
					(a.type				= 'group'			AND 
					 a.magic			= 'all')
				OR
					(a.type				= 'group'			AND
					 a.magic			= 'self'			AND
					 map_u.object_id	= u.object_id		AND
					 map_o.object_id	= ?					AND
					 map_o.group_id		= map_u.group_id	AND
					 u.id				= ?)
			);
	};
	my @bind_params = ($user_id, $user_id, $oid, $oid, $oid, $oid, $user_id);
	defined(dbSelect($sql, @bind_params)) && return 1;
	
	return 0;
}

sub authorizeAccessToGroup
{
	my ($user_id, $gid, $access) = @_;

	my %access_map = (
		read	=> 1,
		write	=> 2,
		admin	=> 4,
	);

	#--------------------------------------------------------------------
	# the magical value 'self' is not allowed for administration of
	# groups.  Shut it off by looking for something impossible.
	#--------------------------------------------------------------------
	my $self = ($access eq 'admin') ? 'junk' : 'self';

	exists($access_map{$access}) || die "access-type [$access] uknown";

	my $implicit_read;
	if ($access eq 'read') {
		$implicit_read = qq{
			OR
				(u.id			= ?			 AND
				 map.object_id	= u.object_id)
		};
	}

	my $sql = qq{
		SELECT a.id AS id
		FROM
			c_access		AS a,
			c_group_object	AS map,
			v_user			AS u
		WHERE
			(a.user_id = ? AND a.level >= $access_map{$access}) AND
			(
					(a.type			= 'group'		AND
					 a.ref_id		= ?)
				OR
					(a.type			= 'group'		AND
					 a.magic		= 'self'		AND
					 map.group_id	= ?				AND
					 map.object_id	= u.object_id	AND
					 u.id			= ?)
				OR
					(a.type			= 'group'		AND 
					 a.magic		= 'all')
				$implicit_read
			);
	};

	my @bind_params = ($user_id, $gid, $gid, $user_id);
	$implicit_read && push(@bind_params, $user_id);

	defined(dbSelect($sql, @bind_params)) && return 1;
	
	return 0;
}

sub findReadableGroups
{
	my $user_id = shift;

	#---------------------------------------------------------------
	# three things need to be checked:
	# 1. find all groups to which the object associated with
	#    $user_id belongs (if any)
	# 2. find access records that refer directly to a particular group
	# 3. check if the user has an access record with type=group, and
	#    magic=all, if so, return *all* groups
	#---------------------------------------------------------------

	my $sql = qq{
		SELECT map.group_id
		FROM
			c_group_object	AS map,
			v_user			AS u
		WHERE
			(u.id			= ? AND
			 map.object_id	= u.object_id)

		UNION

		SELECT ref_id AS group_id
		FROM c_access
		WHERE
			(user_id		= ?) AND
			(type			= 'group') AND 
			(ref_id			IS NOT NULL)

		UNION

		SELECT g.id AS group_id
		FROM
			v_group g,
			c_access a
		WHERE
		 a.user_id = ? AND a.type = 'group' AND a.magic = 'all';
	};
	my @bind_params = ($user_id, $user_id, $user_id);

	my $sth = $DBH->prepare($sql);
	$sth->execute(@bind_params) || confess "statement [$sql] failed";
	my $tbl_ary_ref = $sth->fetchall_arrayref([0]);

	my @result;
	foreach my $row (@{$tbl_ary_ref}) {
		push(@result, $row->[0]);
	}

	@result;
}

sub findAdministratableUsers
{
	my ($user_id, $op) = @_;

	my @all_users = getUsers();

	($AUTH || isSuperuser($user_id)) && return @all_users;

	my @admin_groups = findAdminGroups($user_id);

	my @administratable_users;
	foreach my $uid (@all_users) {
		my @readable = findReadableGroups($uid);
		if (superset(\@admin_groups, \@readable)) {
			push(@administratable_users, $uid);
		}
	}

	return @administratable_users;
}

sub findAdminGroups
{
	my $user_id = shift;

	my $sql = qq{
		SELECT ref_id AS group_id
		FROM c_access AS a
		WHERE
			a.user_id	= ?			AND
			a.level		= 4			AND
			a.type		= 'group'	AND
			a.ref_id	IS NOT NULL

		UNION

		SELECT g.id AS group_id
		FROM
			v_group g,
			c_access a
		WHERE
			a.user_id	= ?			AND
			a.level		= 4			AND
			a.type		= 'group'	AND
			a.magic	= 'all';
	};
	my @bind_params = ($user_id, $user_id);

	#$DEBUG && abslog("findAdminGroups: sql:\n\t".join("\n\t", split(/\n/,$sql)));
	return dbSelectColumn($sql, 'group_id', @bind_params);
}

sub findAdministratableObjects
{
	my ($user_id, $op) = @_;

	my $su = isSuperuser($user_id);

	$su && return getPeople(0);

	# get a list of all objects in all groups administratable by $user_id
	my $sql = qq{
		SELECT DISTINCT go.object_id
		FROM
			c_access		AS a,
			v_object		AS o,
			v_group			AS g,
			c_group_object	AS go
		WHERE
			a.level			= 4				AND
			a.user_id		= ?				AND
			a.type			= 'group'		AND
			a.ref_id		= go.group_id;
	};
	my @objects = dbSelectColumn($sql, 'object_id', $user_id);

	my @new;

	#--------------------------------------------------------
	# to be allowed to delete an object, a group-admin must
	# have admin privs for all groups to which the object belongs
	#--------------------------------------------------------
	if ($op eq 'delete') {
		foreach my $object_id (@objects) {
			if (adminForAllGroups($user_id, $object_id)) {
				push(@new, $object_id);
			}
		}
		@objects = @new;
	}
	
	@new = ();
	#--------------------------------------------------------
	my @super_users = getSuperusers();
	foreach my $oid (@objects) {
		my $o_uid = findUidForObject($oid);
		if (defined($o_uid) && inListN($o_uid, \@super_users)) { next; }
		push(@new, $oid);
	}

	return @new;
}

sub adminForAllGroups
{
	my ($user_id, $object_id) = @_;

	# get groups for $user_id
	my $sql = qq{
		SELECT ref_id
		FROM c_access AS a
		WHERE
			a.user_id	= ?			AND
			a.level		= 4			AND
			a.type		= 'group'	AND
			a.ref_id	IS NOT NULL;
	};
	my @u_groups = dbSelectColumn($sql, 'ref_id', $user_id);

	my @o_groups = getObjectGroups($object_id);

	# is $user_id admin for all groups in which $object_id is member?
	foreach my $o_gid (@o_groups) {
		inListN($o_gid, \@u_groups) || return 0;
	}
	return 1;
}

sub getSuperusers
{
	my $sql = qq{
		SELECT a.user_id
		FROM
			c_access AS a
		WHERE
			a.type	= 'group'	AND
			a.magic	= 'all';
	};
	return dbSelectColumn($sql, 'user_id');
}

sub isSuperuser
{
	my $user_id = shift;

	my $sql = qq{
		SELECT id
		FROM c_access AS a
		WHERE
			a.user_id	= ?			AND
			a.level		= 4			AND
			a.type		= 'group'	AND
			a.magic		= 'all';
	};
	defined(dbSelect($sql, $user_id)) && return 1;
	return 0;
}

sub userHasMagicAccess
{
	my ($uid, $type, $value) = @_;

	my $sql = qq{
		SELECT level
		FROM c_access
		WHERE
			user_id	= ?			AND
			type	= ?			AND
			magic	= ?;
	};

	my $ref = dbSelect($sql, $uid, $type, $value);
	return defined($ref) ? $ref->{level} : undef;
}

sub deleteGidInAcls
{
	my $gid = shift;

	my $sql = qq{
		DELETE FROM c_access
		WHERE ref_id = ?
		AND type = 'group';
	};
	$DBH->do($sql, undef, $gid);
}

sub deletePidInAcls
{
	my $pid = shift;

	my $sql = qq{
		DELETE FROM c_access
		WHERE ref_id = ?
		AND type = 'object';
	};
	$DBH->do($sql, undef, $pid);
}

sub deleteUserIdInAcls
{
	my $uid = shift;

	my $sql = qq{
		DELETE FROM c_access
		WHERE user_id = ?;
	};
	$DBH->do($sql, undef, $uid);
}

sub setAcls
{
	my ($uid, $pacl_ref, $gacl_ref) = @_;

	# get rid of old ACLs
	deleteUserIdInAcls($uid);

	my $sql = qq{
		INSERT INTO c_access
			(user_id, ref_id, type, magic, level)
		VALUES
			($uid, ?, ?, ?, ?);
	};
	my $sth = $DBH->prepare($sql);
	defined($sth) || die "prepare failed on sql [$sql]";

	my @params;
	if (defined($pacl_ref)) {
		foreach my $acl (@{ $pacl_ref }) {
			if ($acl->{target} eq 'self') {
				@params = (undef, 'object', 'self', $acl->{level});
			} elsif ($acl->{target} =~ /^\d+$/) {
				@params = ($acl->{target}, 'object', undef, $acl->{level});
			} else {
				die "target [$acl->{target}] cannot be";
			}
			$sth->execute(@params);
		}
	}

	if (defined($gacl_ref)) {
		foreach my $acl (@{ $gacl_ref }) {
			if ($acl->{target} =~ /^(self|all)$/) {
				@params = (undef, 'group', $acl->{target}, $acl->{level});
			} elsif ($acl->{target} =~ /^\d+$/) {
				@params = ($acl->{target}, 'group', undef, $acl->{level});
			} else {
				die "target [$acl->{target}] cannot be";
			}
			$sth->execute(@params);
		}
	}

	$sth->finish;
}

sub deleteElementN
{
	my($elem, $lref) = @_;

	my $count = 0;
	my $match = 0;
	foreach my $tmp (@{$lref}) {
		if ($tmp == $elem) {
			$match = 1;
			last;
		}
		$count++;
	}

	if ($match) {
		splice(@{$lref}, $count, 1);
	}
}

#--------------------------------------------------------------------
# Group stuff
#--------------------------------------------------------------------

sub getGroups
{
	my $sql = qq{
		SELECT id FROM v_group;
	};
	my $aref = $DBH->selectcol_arrayref($sql);
	return @{$aref};
}

#-------------------------------------------------------------
# getObjectGroups() get all groups to which an object belongs
#-------------------------------------------------------------
sub getObjectGroups
{
	my $oid = shift;

	my $sql = qq{
		SELECT group_id
		FROM c_group_object
		WHERE object_id = ?;
	};

	return dbSelectColumn($sql, 'group_id', $oid);
}

sub getGroup
{
	my ($gid, $field) = @_;
	
	if (!groupExists(id => $gid)) {
		#print "no group with ID=[$gid]\n";
		return undef;
	}
	
	my $sql = qq{
		SELECT *
		FROM v_group
		WHERE id = ?;
	};
	my $ref = dbSelect($sql, $gid);

	if ($field) {
		exists($ref->{$field}) || die "field [$field] unknown in group";
		return $ref->{$field};
	}

	return $ref;
}

sub groupCount
{
	my $sql = qq{SELECT id FROM v_group;};
	my @list = dbSelect($sql);
	return scalar(@list);
}

#--------------------------------------------------------------------
# user stuff
#--------------------------------------------------------------------

sub getUser
{
	my ($thing, $val, $field) = @_;

	my $sql = qq{
		SELECT *
		FROM v_user
		WHERE $thing = ?;
	};
	my $ref = dbSelect($sql, $val);
	defined($ref) || return undef;

	if ($field) {
		exists($ref->{$field}) || die "field [$field] unknown in c_user";
		return $ref->{$field};
	}

	return $ref;
}

#--------------------------------------------------------------------
# getUsers()
# 
# if first param is 'all', then returns a list of all user IDs known
# with no params, returns a list of user IDs for users not associated
# with objects
#--------------------------------------------------------------------
sub getUsers
{
	my $all = shift;

	my $sql;
	if ($all eq 'all') {
		$sql = qq{ SELECT id FROM v_user; };
	} else {
		$sql = qq{
			SELECT v_user.id
			FROM v_user
			WHERE v_user.object_id IS NULL;
		};
	}
	my $aref = $DBH->selectcol_arrayref($sql);
	defined($aref) && return (@{$aref});
	return ();
}

#-------------------------------------------------------------
# getUserGroups() get all groups to which a user's object belongs
#-------------------------------------------------------------
sub getUserGroups
{
	my $uid = shift;

	my $sql = qq{
		SELECT go.group_id
		FROM
			c_group_object	AS go,
			v_user			AS u
		WHERE
			u.id			= ?	AND
			go.object_id	= u.object_id;
	};

	return dbSelectColumn($sql, 'group_id', $uid);
}

# ROBBO: check what callers of setPassword expect to pass as uid...
sub setPassword
{
	my ($uid, $password) = @_;

	$DEBUG && abslog(["changePassword: start: user_id=[$uid]",
		"password=[$password]"]);

	#my $ret = checkLengths('c_user', {
	#	password	=> $password,
	#};
	#$ret && return "baddata: $ret";

	if (!userExists(id => $uid)) {
		return 'disappeared';
	}

	my $new_pw = ($PW_HASH_FORMAT eq 'md5')
		? md5_base64($password)
		: $password;

	if ($CRED_SRC eq 'absence') {
		my $sql = qq{
			UPDATE c_user SET password = ? WHERE id = ?;
		};
		$DBH->do($sql, undef, $new_pw, $uid);
	} elsif ($CRED_SRC eq 'htaccess') {
		# TODO
		my $ret = setHtaccessPassword($uid, $password);
		if ($ret) { return 'error'; }
	} else {
		die "unimplemented method [$CRED_SRC]";
	}

	return 'ok';
}

#--------------------------------------------------------------------
# setHtaccessPassword() is an EXAMPLE of how you could
# implement password management for HTTP authentication.
# Unless you're using SSL, using this method means you
# will be sending cleartext passwords over the network.
# If you do not want to use absence to manage your passwords
# (for example if you are using HTTP Digest authentication
# in order to avoid sending cleartext passwords over the
# network) then you should set "manage_password" in the
# configuration to "no".  You will need to come up with
# your own method of managing passwords.
#--------------------------------------------------------------------
sub setHtaccessPassword
{
	my ($uid, $password) = @_;

	my $username = getUser(id => $uid, 'username');

	my $htaccess = AbsenceConfig::fetch('htaccess_path');
	my @cmd = ('htpasswd', '-b', $htaccess, $username, $password);

	#dbg("running cmd: ".join(' ', @cmd);

	my ($wfh, $rfh);
	my $pid = open3($wfh, $rfh, undef, @cmd);
	close($wfh);
	my $out;
	while(<$rfh>) {
		$out .= $_;
	}
	close($rfh);

	waitpid($pid, 0);
	my $status = $? >> 8;
	chomp($out);

	#dbg("output: [$out]");

	if ($status) {
		abslog("setHtaccessPassword: external command failed. rc=$status, output:\n$out");
	}
	return $status;
}

# ROBBO: modify callers of modifyUser() to pass uid and not username
sub modifyUser
{
	my ($uid, $password, $pacl, $gacl) = @_;

	$DEBUG && abslog(["modifyUser: start: user_id=[$uid]",
		"password=[$password]",
		"pacl=[$pacl]", "gacl=[$gacl]"]);

	#my $ret = checkLengths('c_user', {
	#	password	=> $password,
	#};
	$ret && return "baddata: $ret";

	if (!userExists(id => $uid)) {
		return 'disappeared';
	}
	my $u_ref = getUser('id' => $uid);

	my $new_pw = (($u_ref->{password} ne $password) && ($PW_HASH_FORMAT eq 'md5'))
		? md5_base64($password)
		: $password;

	my $sql;
	if ($CRED_SRC eq 'absence') {
		$sql = qq{
			UPDATE c_user
			SET password = ?
			WHERE id = ?;
		};
	} elsif ($CRED_SRC ne 'htaccess') {
		die "unimplemented method [$CRED_SRC]";
	}

	if ($CRED_SRC eq 'htaccess') {
		# modify user in some ".htpasswd" file
		my $out = setHtaccessPassword($uid, $password);
	}

	setAcls($uid, $pacl, $gacl);

	$DBH->do($sql, undef, $new_pw, $uid);

	abslog([
		"modify-user: username=[$u_ref->{username}]",
		'pacl='.join(',', map { "$_->{level}:$_->{target}" } @{$pacl}),
		'gacl='.join(',', map { "$_->{level}:$_->{target}" } @{$gacl}),
	]);

	return 'ok';
}

#--------------------------------------------------------------------
# addUser()
# 
# add a 'user' entity to the DB, possibly also adding pacl/gacl records
# input:
# $pacl
# 	a reference to an array of hashes of the form:
# 	[
# 		{ level => <level>, oid => <oid> },
# 		{ level => <level>, oid => <oid> },
# 	...
# 	]
# 	where <oid> is an object-ID (integer), or the string 'self',
# 	and <level> is 1, 2, or 4 for read, write, and admin, respectively
# $gacl
# 	a reference to an array of scalars of the form:
# 	[
# 		{ level => <level>, gid => <oid> },
# 		{ level => <level>, gid => <oid> },
# 	...
# 	]
# 	where <gid> is a group-ID, 'self', or 'all' and <level> is
# 	1, 2, or 4, as above
#--------------------------------------------------------------------
sub addUser
{
	my ($username, $password, $pacl, $gacl) = @_;

	$DEBUG && abslog(["addUser: start: user_id=[$username]",
		"password=[$password]",
		'pacl=['.acl2string($pacl).']', 'gacl=['.acl2string($gacl).']']);

	#my $ret = checkLengths('c_user', {
	#	username	=> $username,
	#	password	=> $password,
	#};
	#$ret && return "baddata: $ret";

	if (userExists(username => $username)) {
		return 'duplicate';
	}

	my $uid = _addUser($username, $password, $pacl, $gacl);

	abslog([
		"add-user: username=[$username]",
		#"password=[$password]",
		'pacl='.join(',', map { "$_->{level}:$_->{target}" } @{$pacl}),
		'gacl='.join(',', map { "$_->{level}:$_->{target}" } @{$gacl}),
	]);

	return wantarray ? ('ok', $uid) : 'ok';
}

sub acl2string
{
	my $acl_list = shift;

	return join(',', map { "$_->{level}:$_->{target}" } @{$acl});
}

# ROBBO: check why I used _adduser() inside of adduser()
# answer: because I need it for addPerson too...
sub _addUser
{
	my ($username, $password, $pacl_ref, $gacl_ref, $oid) = @_;

	my @fields = qw(username);
	my @values = ($username);

	if (defined($oid)) {
		push(@fields, 'object_id');
		push(@values, $oid);
	}

	if ($CRED_SRC eq 'absence') {
		my $pw = ($PW_HASH_FORMAT eq 'md5')
			? md5_base64($password)
			: $password;
		push(@fields, 'password');
		push(@values, $pw);
	} elsif ($CRED_SRC ne 'htaccess') {
		die "unimplemented method [$CRED_SRC]";
	}

	my $fields = join(',', @fields);
	my $binds = join(',', map '?', @values);

	my $sql = qq{
		INSERT INTO c_user
		($fields)
		VALUES ($binds);
	};

	my $uid = insertRowRetrieveSequence('c_user', $sql, @values);

	# now handle pacl/gacl
	setAcls($uid, $pacl_ref, $gacl_ref);

	if ($CRED_SRC eq 'htaccess') {
		my $ret = setHtaccessPassword($uid, $password);
		$ret && die "error while setting htaccess pw: $ret";
	}

	return $uid;
}

sub deleteAccessControlForUser
{
	my $user_id = shift;

	my $sql = qq{
		DELETE FROM c_access
		WHERE user_id = ?;
	};
	$DBH->do($sql, undef, $user_id);
}

sub deleteUser
{
	my $uid = shift;

	$DEBUG && abslog("delete-user: user_id=[$uid]");

	_deleteUser($uid);

	return 'ok';
}

sub _deleteUser
{
	my $uid = shift;

	#my $now = time();

	# this must happen first!
	deleteUserIdInAcls($uid);

	my $sql = qq{
		DELETE FROM c_user WHERE id = ?;
	};
	$DBH->do($sql, undef, $uid);
}

#--------------------------------------------------------------------
# person stuff
#--------------------------------------------------------------------
sub addPerson
{
	my ($pid, $name, $username, $email, $gref, $pw, $pacl, $gacl) = @_;

	$DEBUG && abslog(["addPerson: start: pid=[$pid], name=[$name]",
		"username=[$username]",
		"email=[$email]", "groups=[@{$gref}]"]);

	my $verb;
	my $old_pw;
	my $sql;
	my $uid;

	if ($pid eq 'new') {
		#-------------------------------------------------
		# CREATE
		#-------------------------------------------------
		if (objectExists(name => $name)) {
			return 'duplicate-name';
		} elsif ($AUTH && $OAP && userExists(username => $username)) {
			return 'duplicate-uid';
		}
		$verb = 'add-person';

		$sql = qq{
			INSERT INTO c_object
				(name,email)
			VALUES (?,?);
		};

		$DBH->begin_work;
		$pid = insertRowRetrieveSequence('c_object', $sql, $name, $email);
		#$DBH->do($sql, undef, $name, $email);

		my $xact_ok = 1;
		if ($OAP && $AUTH) {
			# RaiseError is set, so if _addUser fails, script
			# will terminate, and trasaction will not be completed
			$uid = _addUser($username, $pw, $pacl, $gacl, $pid);
		}
		addGroupMappings($pid, @{$gref});
		$DBH->commit;
	} else {
		#-------------------------------------------------
		# UPDATE
		#-------------------------------------------------
		if (!objectExists(id => $pid)) {
			die "error: attempt to modify a non-existent person [$pid]";
		}
		$verb = 'modify-person';

		$sql = qq{
			UPDATE c_object
				SET name = '$name', email = '$email'
				WHERE id = ?;
		};

		$DBH->begin_work;
		$DBH->do($sql, undef, $pid);
		if ($OAP && $AUTH) {
			$uid = findUidForObject($pid);
			if (defined($uid)) {
				my $rv = modifyUser($uid, $pw, $pacl, $gacl);
				if ($rv ne 'ok') {
					$DBH->rollback;
					return 'error';
				}
			} else {
				if (userExists(username => $username)) {
					$DBH->rollback;
					return 'duplicate-username';
				}
				$uid = _addUser($username, $pw, $pacl, $gacl, $pid);
			}
		}
		deleteGroupMappings($pid);
		addGroupMappings($pid, @{$gref});
		$DBH->commit;
	}

	abslog(["$verb: pid=[$pid]", "name=[$name]", "username=[$username]",
		"email=[$email]", "groups = [@{$gref}]"]);

	return wantarray ? ('ok', $pid, $uid) : 'ok';
}

sub findUidForObject
{
	my $oid = shift;

	my $sql = qq{
		SELECT id
		FROM v_user
		WHERE object_id = ?;
	};
	my $ref = dbSelect($sql, $oid);
	return defined($ref) ? $ref->{id} : undef;
}

sub deletePerson
{
	my $pid = shift;

	#abslog("deletePerson: deleting id=[$pid]");

	if (!objectExists(id => $pid)) {
		abslog("deletePerson: id=[$pid] not found");
		return 'bad-id';
	}

	_deleteObject($pid);

	$AUTH && deletePidInAcls($pid);

	return 'ok';
}

#----------------------------------------------------------------------
# this is the internal version.
#----------------------------------------------------------------------
sub _deleteObject
{
	my $oid = shift;

	#----------------------------------------------------------
	# delete all mappings for object
	#----------------------------------------------------------
	my $sql = qq{
		DELETE FROM c_group_object
		WHERE object_id = $oid;
	};
	$DBH->do($sql);

	#----------------------------------------------------------
	# invalidate all reservations of this object
	# well, actually, just delete the row.  someday I may use
	# invalidation
	#----------------------------------------------------------
	#my $now = time();
	#$sql = qq{
	#	UPDATE d_reservation
	#	SET invalidated = int2tsz($now)
	#	WHERE object_id = $oid;
	#};
	$sql = qq{
		DELETE FROM d_reservation
		WHERE object_id = ?;
	};
	$DBH->do($sql, undef, $oid);

	#----------------------------------------------------------
	# remember user-id for later
	#----------------------------------------------------------
	my $user_id = findUidForObject($oid);

	#----------------------------------------------------------
	# remember name for logging
	#----------------------------------------------------------
	my $name = getPerson($oid, 'name');
	abslog("deleting object: oid=[$oid], name=[$name]");

	#----------------------------------------------------------
	# delete associated user, if necessary
	# must be done *before* object is deleted
	#----------------------------------------------------------
	if ($OAP && $AUTH) {
		if (!defined($user_id)) {
			abslog("_deleteObject, deleted object [$oid], auth enabled, but no user_id found for that object");
		}
		_deleteUser($user_id);
	}

	#----------------------------------------------------------
	# delete row.  someday I may invalidate...
	# 
	#----------------------------------------------------------
	#$sql = qq{
	#	UPDATE c_object
	#	SET invalidated = int2tsz($now)
	#	WHERE id = $oid;
	#};
	#----------------------------------------------------------

	$sql = qq{
		DELETE FROM c_object
		WHERE id = ?;
	};
	$DBH->do($sql, undef, $oid);
}

#----------------------------------------------------------------------
# addReservation()
#
# adds or modifies a reservation for a person.
# $start and $end are dates in the form
#	{ day => $day, month => $month, year => $year }
#
#----------------------------------------------------------------------


sub addReservation
{
	my ($res_id, $person_id, $start, $end, $type_id, $desc) = @_;

	my $s = "$start->{year}-$start->{month}-$start->{day}";
	my $e = "$end->{year}-$end->{month}-$end->{day}";

	if ($DEBUG) {
		abslog("addReservation: rid=$res_id, pid=$person_id, start=$s, end=$e, type_id=$type_id, desc=[$desc]");
	}

	my $new = 0;
	my $verb;
	my $sql;
	my ($old_s, $old_e);

	if ($res_id eq 'new') {
		$new = 1;
		$verb = 'add-reservation';
	} else {
		my $ref = getReservation($res_id);
		if (!defined($ref)) {
			abslog("error: attempt to modify non-existent reservation [$res_id]");
			return ('disappeared');
		}
		# a reservation is being modified, I need the old start/end dates
		$old_s = $ref->{start};
		$old_e = $ref->{finish};
		$verb = 'modify-reservation';
		$DEBUG && abslog("old: start=$ref->{start}, finish=$ref->{finish}");
	}

	#----------------------------------------------------------
	# check if either starting or ending day is past end of
	# corresponding month...
	#----------------------------------------------------------
	if (($start->{day} > AbsenceDate::daysInMonth($start->{month})) ||
		($end->{day} > AbsenceDate::daysInMonth($end->{month})))
	{
		return ('badday');
	}

	my $sjd = AbsenceDate::julianDay($start);
	my $ejd = AbsenceDate::julianDay($end);
	($sjd > $ejd) && return ('impossible');

	#my ($ret, $id) = resConflict($res_id, $person_id, $start, $end, $type_id);
	my $ret = resConflict($res_id, $person_id, $start, $end, $type_id);

	#($ret eq 'bad') && return ('conflict', $id);

	if (defined($ret)) {
		abslog("resConflict found conflict");
		return ('conflict', $ret);
	}

	abslog(
		[
			"$verb: res=[$res_id]",
			"person-id = [$person_id]",
			"start = [$s]",
			"end = [$e]",
			"type = [$type_id]",
			"desc = [$desc]",
		]
	);

	_addReservation($res_id, $person_id, $start, $end, $type_id, $desc);

	my @o_groups = getObjectGroups($person_id);

	if (!$new) {
		my $s = convToAbsenceDate($old_s);
		my $e = convToAbsenceDate($old_e);
		updateModificationTimes(\@o_groups, $s, $e);
	}
	updateModificationTimes(\@o_groups, $start, $end);

	$DEBUG && abslog("addReservation: success");

	return ('ok');
}

#----------------------------------------------------------------------
# updateModificationTimes()
#
# parameters:
#	$s, $e references to dates in the form:
#		[<day>, <month>, <year>]
#----------------------------------------------------------------------

sub updateModificationTimes
{
	# ROBBO TODO: check format of $s and $e

	my ($gref, $s, $e) = @_;

	if ($DEBUG) {
		my $tmp = join(',', @{$gref});
		my $start = "$s->{month}/$s->{year}";
		my $end = "$e->{month}/$s->{year}";
		abslog("updateModificationTimes: grps: [$tmp], s: [$start], e: [$end]");
	}

	# modify in-memory hash
	my $month = $s->{month};
	my $year = $s->{year};

	my $end = $e->{year} * 12 + $e->{month};

	my $now = time();

	while(($year*12+$month) <= $end) {
		$DEBUG && abslog("updModTime: updating y=$year, m=$month");
		foreach my $gid (@{$gref}) {
			updateMonthModTime($gid, $month, $year);
		}
		if ($month == 12) {
			$year++;
			$month = 1;
		} else {
			$month++;
		}
	}
}

sub updateMonthModTime
{
	my ($gid, $month, $year) = @_;

	my $now = time();

	my $update_sql = qq{
		UPDATE d_group_mod_time
		SET mod_time = $now
		WHERE group_id = ?
		AND month = ?
		AND year = ?;
	};
	my $rv = $DBH->do($update_sql, undef, $gid, $month, $year);
	defined($rv) || die "update mmt, rv=undef, sql=[$update_sql]";
	if ($rv == 0) {
		my $insert_sql = qq{
			INSERT INTO d_group_mod_time
			(group_id,month,year,mod_time)
			VALUES (?,?,?,$now);
		};
		$rv = $DBH->do($insert_sql, undef, $gid, $month, $year);
		defined($rv) || die "insert mmt, rv=undef, sql=[$insert_sql]";
	}
}

sub updateImageDimensions
{
	my ($gid, $month, $year, $width, $height) = @_;

	my $update_sql = qq{
		UPDATE d_group_mod_time
		SET image_width = $width, image_height = $height
		WHERE group_id = ?
		AND month = ?
		AND year = ?;
	};
	my $rv = $DBH->do($update_sql, undef, $gid, $month, $year);
	defined($rv) || die "update mmt, rv=undef, sql=[$update_sql]";
	if ($rv == 0) {
		my $insert_sql = qq{
			INSERT INTO d_group_mod_time
			(group_id,month,year,image_width,image_height)
			VALUES (?,?,?,$width,$height);
		};
		$rv = $DBH->do($insert_sql, undef, $gid, $month, $year);
		defined($rv) || die "insert mmt, rv=undef, sql=[$insert_sql]";
	}
}

#------------------------------------------------------------------
# updateMonthModTimes()
#
# is passed the group-id and a structure of the form:
#	$ref = {
#		<year1> = {
#			<month1> = <time>,
#			<month2> = <time>,
#		},
#		<year2> = {
#			<month1> = <time>,
#			<month2> = <time>,
#		},
#	}
#------------------------------------------------------------------
sub updateMonthModTimes
{
	my ($gid, $mref) = @_;

	foreach my $year (keys(%{$mref})) {
		foreach my $month (keys(%{$mref->{$year}})) {
			updateMonthModTime($gid, $month, $year);
			$DEBUG && abslog("updating mmt for [gid=$gid,$month/$year]");
		}
	}
}

sub getModificationTime
{
	my ($gid, $month, $year) = @_;

	$DEBUG && abslog("getModificationTime: lookup: gid=$gid, m=$month, y=$year");

	my $sql = qq{
		SELECT mod_time
		FROM d_group_mod_time
		WHERE group_id = ?
		AND month = ?
		AND year = ?;
	};
	my $ref = dbSelect($sql, $gid, $month, $year);

	if (defined($ref)) {
		return $ref->{mod_time};
	} else {
		$DEBUG && abslog("getModificationTime: returning [undef]");
		return undef;
	}
}

sub getImageDimensions
{
	my ($gid, $month, $year) = @_;

	$DEBUG && abslog("getImageDimensions: lookup: gid=$gid, m=$month, y=$year");

	my $sql = qq{
		SELECT image_width, image_height
		FROM d_group_mod_time
		WHERE group_id = ?
		AND month = ?
		AND year = ?;
	};
	my $ref = dbSelect($sql, $gid, $month, $year);

	if (defined($ref)) {
		return ($ref->{image_width}, $ref->{image_height});
	} else {
		$DEBUG && abslog("getImageDimensions: returning [undef]");
		return undef;
	}
}

sub deleteReservation
{
	my $res_id = shift;

	my $res = getReservation($res_id);

	if (!defined($res)) {
		# someone else got there first
		abslog("deleteReservation: ID [$res_id] disappeared (overlapping ops)");
		return 'conflict';
	}

	my ($start, $end) = ($res->{start}, $res->{finish});

	my $pid = $res->{object_id};
	#invalidateReservations(object_id => $res_id);
	deleteReservations(id => $res_id);

	# piggy-back onto DB-lock
	my @o_groups = AbsenceDB::getObjectGroups($pid);
	my $s = convToAbsenceDate($res->{start});
	my $e = convToAbsenceDate($res->{finish});
	updateModificationTimes(\@o_groups, $s, $e);

	abslog("deleteReservation: res-id=[$res_id]");

	return 'ok';
}

#----------------------------------------------------------------
# convToDbDate()
#
# input:
# 	a reference to an array with form [ $day, $month, $year ]
#
# returns:
# 	a string of form "YYYY-MM-DD"
#----------------------------------------------------------------
sub convToDbDate($)
{
	#return join('-', reverse(@{ $_[0] }));
	my $r = shift;
	return join('-', $r->{year}, $r->{month}, $r->{day});
}

#----------------------------------------------------------------
# convToAbsenceDate()
#
# does reverse of convToDbDate()
#----------------------------------------------------------------
sub convToAbsenceDate($)
{
	my ($year, $mon, $day) = split('-', $_[0]);
	#return ( $day, $mon, $year );
	return { day => int($day), month => int($mon), year => $year };
}

sub resConflict
{
	my ($res_id, $pid, $start, $end, $type_id) = @_;

	#abslog("resConflict: res-id=[$res_id]");

	if (!$MULTI_RES) {
		my $db_start = convToDbDate($start);
		my $db_finish = convToDbDate($end);

		my @bind_params = ($pid);
		my $avoid_existing_res;
		if ($res_id =~ /^\d+$/) {
			$avoid_existing_res = qq{(id != ?) AND};
			push(@bind_params, $res_id);
		}
		my $sql = qq{
			SELECT id
				FROM v_reservation
			WHERE
				(object_id = ?) AND $avoid_existing_res (
					(start  >= ? AND start <= ?) OR
					(finish >= ? AND start <= ?) OR
					(start  <  ? AND finish > ?)
			);
		};
		my @tmp = ($db_start, $db_finish);
		push(@bind_params, @tmp, @tmp, @tmp);

		my $ref = dbSelect($sql, @bind_params);
		#$DEBUG && abslog("SQL: $sql");

		return defined($ref) ? { type => 'simple', res => $ref } : undef;
	}

	#------------------------------------------------------------
	# multiple reservations are enabled
	#------------------------------------------------------------

	my ($month, $year) = ($start->{month}, $start->{year});
	my $jm_end = $end->{year} * 12 + $end->{month};
	my $jm;
	my $conflict;
	do {
		my ($sd, $ed)
			= getMonthBounds($start, $end, $month, $year);
		my $myres = {
			res		=> { id => $res_id, type_id => $type_id },
			bounds	=> { start => $sd, end => $ed },
		};
		$jm = $year * 12 + $month;
		my @mres_list = getMonthReservations($pid, $month, $year);
		my @orig_list = @mres_list;

		#----------------------------------------------------------
		# if I am modifying a reservation, I need to remove the old
		# reservation from the list to check
		#----------------------------------------------------------
		if ($res_id ne 'new') {
			my @tmp;
			foreach my $mres (@mres_list) {
				if ($mres->{res}->{id} != $res_id) {
					push(@tmp, $mres);
					#abslog("  adding id=[$mres->{res}->{id}] to list to check");
				}
			}
			@mres_list = @tmp;
		}

		push(@mres_list, $myres);
		my $layout = AbsenceImage::layoutObjectRow(\@mres_list, $month, $year);
		#---------------------------------------------------------
		# check if row would grow beyond allowed boundary
		#---------------------------------------------------------
		my $max_slot = $MAX_MULTI * $FULL_SLOTS - 1;
		if ($layout->{last_slot} > $max_slot) {
			# must check if result would be same without new res
			my $lo2 = AbsenceImage::layoutObjectRow(\@orig_list, $month, $year);
			if ($lo2->{last_slot} != $layout->{last_slot}) {
				#$conflict = { type => 'height' };
				#last;
				return {
					type	=> 'no_space',
					month	=> $month,
					year	=> $year,
				};
			}
		}

		#---------------------------------------------------------
		# check if there is any illegal coincidence
		#---------------------------------------------------------
		my $type_class = getTypeClass($type_id);

		if (defined($layout->{error})) {
			foreach my $err (@{ $layout->{error} }) {
				if ($type_class eq 'block') {
					if ($err->{type} eq 'b_coincidence' &&
						($err->{ids}->[0] eq $res_id ||
						$err->{ids}->[1] eq $res_id))
					{
						return $err;
					}
				} else {
					if ($err->{type} eq 'nb_coincidence' &&
						($err->{ids}->[0] eq $res_id ||
						$err->{ids}->[1] eq $res_id))
					{
						return $err;
					}
				}
			}
		}

		{
			$conflict = 1;
			$jm = $jm_end;
		}
		if ($month == 12) {
			$month = 1; $year++;
		} else {
			$month++;
		}
	} while ($jm < $jm_end);

	return undef;
}

sub _addReservation
{
	my ($res_id, $person_id, $start, $end, $type_id, $desc) = @_;

	$DEBUG && abslog("_addReservation: res_id = [$res_id].");

	my $s = convToDbDate($start);
	my $e = convToDbDate($end);

	my $sql;
	my @bind_params;
	if ($res_id eq 'new') {
		$sql = qq{
			INSERT INTO d_reservation
			(object_id,type_id,start,finish,description)
			VALUES (?,?,?,?,?);
		};
		@bind_params = ($person_id,$type_id,$s,$e,$desc);
	} else {
		$sql = qq{
			UPDATE d_reservation
			SET
				start		= ?,
				finish		= ?,
				type_id		= ?,
				description	= ?
			WHERE id = ?;
		};
		@bind_params = ($s, $e, $type_id, $desc, $res_id);
	}

	return $DBH->do($sql, undef, @bind_params);
}

sub timestamp
{
	my @months = qw(Jan Feb Mar Apr May Jun Jul Aug Sep Oct Nov Dec);
	my ($sec,$min,$hour,$mday,$mon,$year,@junk) = localtime(time);

	return sprintf("%d-%s-%d %02d:%02d",
		$mday, $months[$mon], $year+1900, $hour, $min);
}

sub dbg
{
	my $msg = shift;

	$DEBUG && abslog("DEBUG: $msg");
}

#===========================================================================
#-------------------------------------------------------------------
# dbSelect()
#
# performs a generic query using fetchrow_hashref and returns:
#
#       (array context) references to the row-hashes
#       (scalar context) reference to first row
#
#-------------------------------------------------------------------
sub dbSelectColumn
{
	my ($sql, $col, @bind_params) = @_;

	my $sth;
	eval {
		$sth = $DBH->prepare($sql);
	};
	if ($@) {
		confess $@;
	}
	defined($sth) || die "prepare failed on sql [$sql]";

	$sth->execute(@bind_params) || confess "statement [$sql] failed";

	my @result;
	while(defined($ref = $sth->fetchrow_hashref)) {
		exists($ref->{$col}) || die "column [$col] unknown in sql [$sql]";
		push(@result, $ref->{$col});
	}

	@result;
}

sub dbSelect
{
	my ($sql, @bind_params) = @_;

	my $sth;
	eval {
		$sth = $DBH->prepare($sql);
	};
	if ($@) {
		confess $@;
	}
	defined($sth) || die "prepare failed on sql [$sql]";

	$sth->execute(@bind_params) || confess "statement [$sql] failed";

	my @result;
	my $ref;
	while(defined($ref = $sth->fetchrow_hashref)) {
		push(@result, $ref);
	}

	if (wantarray) {
		return @result;
	} else {
		@result || return undef;
		return $result[0];
	}
}

sub executeStatement
{
	my ($sql, @bind_params) = shift;

	$DEBUG && abslog("   execing statement: $sql");
	$FAKE && return 1;

	my $rv;
	eval {
		$rv = $DBH->do($sql, undef, @bind_params);
	};
	if ($@) {
		confess "SQL statement failed: [$sql]";
	}

	#if (!$rv) {
	#	confess "statement [$sql] failed";
	#}

	defined($rv) || return undef;
	($rv eq '0E0') && return 0;
	return $rv;
}

sub insertRowRetrieveSequence
{
	my ($table, $sql, @bind_params) = @_;

	$DEBUG && abslog("   insertRowRetrieveSequence: $sql");
	$FAKE && return 1;

	my ($rv, $sth);

	eval {
		$sth = $DBH->prepare($sql);
	};
	if ($@) {
		confess "SQL prepare failed: [$sql]";
	}
	defined($sth) || die "prepare failed on sql [$sql]";

	$sth->execute(@bind_params) || confess "statement [$sql] failed";

	$sth->finish;

	#----------------------------------------------------------
	# get sequence number
	#----------------------------------------------------------
	my $sequence = "${table}_id_seq";
	$sth = $DBH->prepare("SELECT currval('$sequence');");
	$sth->execute || confess "statement [SELECT currval('$sequence');] failed";

	my $id = ($sth->fetchrow_array)[0];
	defined($id) || confess "no ID returned.";
	return $id;
}

#=======================================================================
# Absence-Type stuff
#=======================================================================
# ROBBO
sub getTypes
{
	my $arg = shift;

	my $sql;
	if ($arg eq 'invalid') {
		$sql = qq{
			SELECT * FROM c_reservation_type WHERE invalidated IS NOT NULL;
		};
	} elsif ($arg eq 'all') {
		$sql = qq{
			SELECT * FROM c_reservation_type;
		};
	} else {
		$sql = qq{
			SELECT * FROM v_reservation_type;
		};
	}

	my @list = dbSelect($sql);
}

sub getType
{
	my $type_id = shift;

	my $sql = qq{ SELECT * FROM c_reservation_type WHERE id = ?; };
	return dbSelect($sql, $type_id);
}

sub getTypeClass
{
	my $type_id = shift;

	my $ref = getType($type_id);
	return ($ref->{height} eq 'block') ? 'block' : 'non_block';
}


sub queryTypeName
{
	my ($name, $arg, $ignore) = @_;

	my $table = ($arg eq 'invalid')
		? 'c_reservation_type'
		: 'v_reservation_type';

	my $sql;
	my @bind_params;
	if (defined($ignore)) {
		$sql = qq{ SELECT id FROM $table WHERE name = ? AND id != ?; };
		@bind_params = ($name, $ignore);
	} else {
		$sql = qq{ SELECT id FROM $table WHERE name = ?; };
		@bind_params = ($name);
	}

	return dbSelect($sql, @bind_params);
}

sub modifyType
{
	my $ref = shift;

	my $sql;

	$DBH->begin_work;
	if ($ref->{default_type}) {
		# first 
		$sql = qq{
			UPDATE c_reservation_type
			SET default_type = FALSE
			WHERE default_type = TRUE;
		};
		if (!defined($DBH->do($sql))) {
			die "modifyType (remove default) failed";
		}
	}

	my @pairs;
	my @bind_params;
	foreach my $key (keys(%{ $ref })) {
		next if ($key eq 'id');
		push(@pairs, "$key = ?");
		push(@bind_params, $ref->{$key});
	}
	my $pairs = join(',', @pairs);
	$sql = qq{
		UPDATE c_reservation_type
		SET $pairs
		WHERE id = $ref->{id};
	};
	if (!defined($DBH->do($sql, undef, @bind_params))) {
		$DBH->rollback;
		die "modifyType (modify) failed";
	}
	setTypeModTime();
	$DBH->commit;
}

sub addType
{
	my $ref = shift;

	my (@fields, @values);

	foreach my $key (keys(%{ $ref })) {
		push(@fields, $key);
		push(@values, $ref->{$key});
	}
	my $fields = join(',', @fields);
	my $binds = join(',', map '?', @values);

	my $sql = qq{
		INSERT INTO c_reservation_type
		($fields)
		VALUES ($binds);
	};
	my $tid = insertRowRetrieveSequence('c_reservation_type', $sql, @values);

	return $tid;
}

sub deactivateType
{
	my $type_id = shift;

	my $now = time();
	my $sql = qq{
		UPDATE c_reservation_type
		SET invalidated = int2tsz($now)
		WHERE id = ?;
	};
	$DBH->do($sql, undef, $type_id);
}

sub reactivateType
{
	my $type_id = shift;

	my $sql = qq{
		UPDATE c_reservation_type
		SET invalidated = NULL
		WHERE id = ?;
	};
	$DBH->do($sql, undef, $type_id);
}


sub getTypeColor
{
	my $type_id = shift;

	my $sql = qq{ SELECT * FROM v_reservation_type WHERE id = ?; };
	my $type_ref = dbSelect($sql, $type_id);
	defined($type_ref) || die "no type found with id=[$type_id]";

	return [
		$type_ref->{color_red},
		$type_ref->{color_green},
		$type_ref->{color_blue},
	];
}

sub getDefaultType
{
	my $sql = qq{ SELECT * FROM v_reservation_type WHERE default_type = TRUE; };
	my $type_ref = dbSelect($sql);
	defined($type_ref) || return undef;

	return $type_ref;
}

sub getTypeModTime
{
	my $config_name = 'type-modification-time';

	return getConfigItem($config_name);
}

sub setTypeModTime
{
	my $config_name = 'type-modification-time';

	setConfigItem($config_name, time());
}

sub resTypeCoincidenceAllowed
{
	my ($type1, $type2) = @_;

	my @ids = ($type1 > $type2) ? ($type2, $type1) : ($type1, $type2);

	my $sql = qq{
		SELECT id
		FROM c_no_coincidence
		WHERE type_id1 = ? AND type_id2 = ?;
	};

	defined(dbSelect($sql, @ids)) && return 0;

	return 1;
}

sub checkVorbidTypeCoincidence
{
	my ($type1, $type2) = @_;

	my @ids = ($type1 > $type2) ? ($type2, $type1) : ($type1, $type2);

	my $sql = qq{
		SELECT id
		FROM c_no_coincidence
		WHERE type_id1 = ? AND type_id2 = ?;
	};

	return defined(dbSelect($sql, @ids)) ? 1 : 0;
}

sub vorbidTypeCoincidence
{
	my ($type1, $type2) = @_;

	my @ids = ($type1 > $type2) ? ($type2, $type1) : ($type1, $type2);

	my $sql = qq{
		INSERT INTO c_no_coincidence
		(type_id1, type_id2) VALUES (?,?);
	};

	$DBH->do($sql, undef, @ids);
}

sub purgeTypeCoincidenceEntries
{
	my $type_id = shift;

	my $sql = qq{
		DELETE FROM c_no_coincidence
		WHERE type_id1 = ?
		OR type_id2 = ?;
	};

	$DBH->do($sql, undef, $type_id, $type_id);
}

sub getTypeCoincidence
{
	my $type_id = shift;

	my $sql = qq{
		SELECT *
		FROM c_no_coincidence
		WHERE type_id1 = ?
		OR type_id2 = ?;
	};

	my @list = dbSelect($sql, $type_id, $type_id);
	@list || return ();

	my @out;
	foreach my $row (@list) {
		push(@out, ($row->{type_id1} == $type_id)
			? $row->{type_id2}
			: $row->{type_id1});
	}

	return @out;
}


#=======================================================================
# Absence-Holiday stuff
#=======================================================================

sub getHolidayList
{
	my ($month, $year, %args) = @_;

	my $dim = AbsenceDate::daysInMonth($month, $year);
	my $month_begin = "$year-$month-1";
	my $month_end = "$year-$month-$dim";

	my $addit = '';
	my @addit;
	my @addit_binds;
	
	if (exists($args{region})) {
		push(@addit, "region_id = ?");
		push(@addit_binds, $args{region});
	}
	if (exists($args{country})) {
		push(@addit, "country_id = ?");
		push(@addit_binds, $args{country});
	}
	if (@addit) {
		$addit = 'AND (' . join(' OR ', @addit) . ')';
	}

	my $sql = qq{
		SELECT * FROM c_holiday
		WHERE
			(day BETWEEN ? AND ?
			OR day BETWEEN ? AND ?)
			$addit
		ORDER BY EXTRACT(DAY FROM day);
	};
	#abslog("%%% getHolidayList: SQL: [$sql]");

	# make sure last day in february, 0001 is 28
	if ($month == 2) { $dim = 28; }

	my @bind_params = (
		$month_begin, $month_end, "0001-$month-01", "0001-$month-$dim",
		@addit_binds
	);
	my @list = dbSelect($sql, @bind_params);
	my %unique;

	my @out;
	#-----------------------------------------------------
	# eliminate duplicate days (should not happen...) and
	# format the way the pre-DB version did: a sorted array
	# containing refs to
	# 	[ $day, $month, $year, $description ]
	# (sorting is done in SQL query!)
	#-----------------------------------------------------
	foreach my $date (@list) {
		my $dref = convToAbsenceDate($date->{day});
		exists($unique{ $dref->{day} }) && next;
		push(@out, [ $dref->{day}, $dref->{month}, $dref->{year}, $date->{description} ]);
		$unique{ $dref->{day} } = 1;
	}

	return @out;
}

sub getRegionId
{
	my $region_name = shift;

	my $sql = qq{
		SELECT id
		FROM c_region
		WHERE name = ?;
	};

	my $ref = dbSelect($sql, $region_name);
	return defined($ref) ? $ref->{id} : undef;
}

sub getCountryId
{
	my ($field, $value) = @_;

	my $sql = qq{
		SELECT id
		FROM c_country
		WHERE $field = ?;
	};

	my $ref = dbSelect($sql, $value);
	return defined($ref) ? $ref->{id} : undef;
}

#=======================================================================
# config-items
#=======================================================================

sub getConfigItem
{
	my $name = shift;

	my $sql = qq{ SELECT * FROM c_config WHERE name = ?; };
	my $config_item = dbSelect($sql, $name);
	defined($config_item) || return undef;

	return $config_item->{content};
}

sub setConfigItem
{
	my ($name, $value) = @_;

	my $sql = qq{
		UPDATE c_config SET content = ? WHERE name = ?;
	};

	my $rv = $DBH->do($sql, undef, $value, $name);
	if ($rv == 0) {
		$sql = qq{
			INSERT INTO c_config (name,content) VALUES (?,?);
		};
		defined($DBH->do($sql, undef, $name, $value))
			|| die "insert c_config, rv=undef, sql=[$sql]";
	}
}

#=======================================================================
# session functions
#=======================================================================

sub checkSession
{
}

sub getSession
{
	my ($sid, $create) = @_;

	$SESSION_PARAMS[1] = $sid;

	my $session = CGI::Session->load(@SESSION_PARAMS)
		or die CGI::Session->errstr;

	if ( $session->is_expired ) {
		$DEBUG && abslog("sess expired");
		$session->delete;
		$create || return undef;

		$session = $session->new(@SESSION_PARAMS);
		$session->param(authed => 0);
		$session->flush;
	} elsif ( $session->is_empty ) {
		$create || return undef;
		$session = $session->new(@SESSION_PARAMS);
		my $id = $session->id();
		$DEBUG && abslog("session empty. created new ID=[$id], authed=0");
		$session->param(authed => 0);
		$session->flush;
	} else {
		my $id = $session->id();
		my $authed = $session->param('authed');
		$DEBUG && abslog("session still good, ID=[$id], authed=[$authed]");
	}

	return $session;
}

#=======================================================================
# old session functions
#=======================================================================

sub old_checkSession
{
	my ($user_id, $ip_addr) = @_;

	my $sql = qq{
		SELECT id, tsz2int(t_stamp)
		FROM d_session
		WHERE user_id = ?
		AND ip_addr = ?;
	};

	# get record, if exists
	my $ref = dbSelect($sql, $user_id, $ip_addr);
	defined($ref) || return 'timeout';
	
	if ($SESSION_TIMEOUT == 0) { return 'ok'; }

	my $esecs = time();
	$DEBUG && abslog("checkSession: now=[$esecs], db_time=[$ref->{tsz2int}]");
	# check if timestamp is recent
	if (($esecs - $ref->{tsz2int}) > $SESSION_TIMEOUT * 60) {
		return 'timeout';
	}

	# timestamp, is recent, so update it
	$sql = qq{
		UPDATE d_session
		SET t_stamp = int2tsz($esecs)
		WHERE id = $ref->{id};
	};
	$DBH->do($sql);

	return 'ok';
}

sub addSession
{
	my ($user_id, $ip_addr, $id) = @_;

	$DEBUG && abslog("addSession()\ntrying UPDATE");
	my $esecs = time();
	my @bind_params;
	my $sql = qq{
		UPDATE d_session
		SET t_stamp = int2tsz($esecs)
	};

	if (defined($id)) {
		$sql .= "WHERE id = ?;";
		@bind_params = ($ref->{id});
	} else {
		$sql .= "WHERE user_id = ? AND ip_addr = ?;";
		@bind_params = ($user_id, $ip_addr);
	}

	my $rv = $DBH->do($sql, undef, @bind_params);
	$DEBUG && abslog("sql=[$sql], rv=[$rv]");
	defined($rv) || die "update session, rv=undef, sql=[$sql]";
	if ($rv == 0) {
		$sql = qq{
			INSERT INTO d_session
			(user_id,ip_addr,t_stamp)
			VALUES (?,?, int2tsz($esecs));
		};
		$rv = $DBH->do($sql, undef, $user_id, $ip_addr);
		$DEBUG && abslog("trying INSERT. sql=[$sql]\nrv=[$rv]");
		$rv || die "insert into d_session failed. user_id=$user_id, ip=$ip_addr, ts=$esecs";
	}
}

sub deleteSession
{
	my ($user_id, $ip_addr, $id) = @_;

	my $sql = qq{
		DELETE FROM d_session
		WHERE user_id = ?
		AND ip_addr = ?;
	};
	$DBH->do($sql, undef, $user_id, $ip_addr);
}

#=======================================================================
# misc stuff
#=======================================================================

sub superset
{
	my ($gref1, $gref2) = @_;

	my $count = 0;
	foreach my $mem (@{$gref2}) {
		if (inListN($mem, $gref1)) {
			$count++;
		}
	}

	dbg("superset: count=[$count], len(gref2) = [".@{$gref2}."]");
	($count == @{$gref2}) && return 1;
	return 0;
}

sub checkLengths
{
	my ($table, $field_ref) = @_;

	my $out;
	foreach my $field (%{ $field_ref }) {
		my $content = $field_ref->{$field};
		my $c_len = defined($content) ? length($content) : 0;
		my $f_len = getFieldLength($table, $field);
		defined($f_len) || next;
		if ($c_len > $f_len) {
			$out .= "field [$field] too long (max=$f_len).<BR>";
		}
	}

	return $out;
}

#=======================================================================
# PostgreSQL specific stuff...
#=======================================================================

sub getFieldDatatype
{
	my ($table, $field) = @_;

	if (!exists($META{$table})) {
		getTableMetadata($table);
	}

	exists($META{$table}->{$field}) ||
		confess "field [$field] not in table [$table]";

	return $META{$table}->{$field}->{type};
}

sub getFieldLength
{
	my ($table, $field) = @_;

	if (!exists($META{$table})) {
		getTableMetadata($table);
	}

	exists($META{$table}->{$field}) ||
		die "field [$field] not in table [$table]";

	return $META{$table}->{$field}->{length};
}

sub getTableMetadata
{
	my $table = shift;

	my $sql = qq{
		SELECT pg_attribute.attname,pg_attribute.atttypmod,pg_type.typname
		FROM pg_class,pg_attribute,pg_type
		WHERE pg_attribute.attnum > 0
		AND pg_attribute.attrelid = pg_class.oid
		AND pg_class.relname = '$table'
		AND pg_attribute.atttypid = pg_type.oid;
	};

	my @list = dbSelect($sql);
	@list || die "table [$table] not found";

	my $length;
	foreach my $ref (@list) {
		# get length. totally non-portable.  pg_attribute.atttypmod stores
		# the length (presumably) including internal length information
		# of 4 bytes, which must be subtracted
		$length = ($ref->{typname} eq 'varchar')
			? $length = $ref->{atttypmod} - 4
			: undef;
		$META{$table}->{$ref->{attname}} = {
			type	=> $ref->{typname},
			length	=> $length,
		};
	}
}

1;

__END__

=head1 NAME

AbsenceDB.pm - interface to absence database.

=head1 WARNING

this documentation isn't necessarily accurate.  If you need proper
documenation, get in touch with me, and I'll make more of an effort.
-RLU

=head1 SYNOPSIS

  use AbsenceDB;


Operations on objects (people):

  @people_ids = getPeople($group_id);
  
  $ref        = getPerson($person_id);
	or
  $thing      = getPerson($person_id, 'field-name');

  $ref        = getGroup($group_id);
	or
  $thing      = getGroup($group_id, 'field-name');

  $result     = addPerson($person_id, $name, $username, $email,
                $group_ref, $password, $pacl, $gacl);

  $result     = deletePerson($person_id);

  $uid        = findUidForObject($oid);


Operations on reservations:

  @list    = getReservations($person_id);

  @list    = getMonthReservations($person_id, $month, $year);

  $res_ref = getReservation($reservation_id);

  $result  = addReservation($reservation_id, $start, $end, $type, $description);

  $result  = deleteReservation($reservation_id);


Operations on groups:

  $result    = addGroup($group_id, $group_name, $group_password, $description);

  $result    = deleteGroup($group_id);

  @group_ids = getGroups();

  $ref       = getGroup($group_id);
	or
  $thing     = getGroup($group_id, 'field-name');

  $number_of_groups = groupCount();

  $mod_time = getModificationTime($group_id, $month, $year);

  $result    = groupExists('field', 'value');

  addGroupMappings($object_id, @group_ids);
  deleteGroupMappings($object_id, @group_ids);
  removeObjectFromGroup(-pid => $object_id, -gid => $group_id);

=head1 DESCRIPTION

C<AbsenceDB.pm> is a crude interface to a database.  In theory, the
underlying database could be anything.  You have perhaps noticed that
elsewhere I refer to B<absences>, but here they are called
B<reservations>?  That came about as a result of my original intention
to keep the system generic, i.e., not necessarily only for people.
Along the way I apparently got a little confused.

=head1 PERSON FUNCTIONS

=over 4

=item getPeople()

returns a list of person_ids belonging to the specified group.

=item getPerson()

retrieves information about the specified person_id.  Without additional
parameters, returns a ref to a object-row from the DB.  With an optional
field-name as parameter, returns the value of that field.

=item addPerson()

creates a new person or modifies an existing one.  Parameters are
C<($person_id, $name, $email, $group_id)>.  If B<$person_id> equals
'new', a new person is created, otherwise B<$person_id> must
contain a valid person-id.  If an attempt is made to create a person
with the same name as an existing one, the operation fails and
the return-value is 'duplicate'.  On success returns 'ok'.

=item deletePerson()

guess what it does? Takes as single param a person-id.  Returns
'ok' on success, otherwise 'bad-id' if the person-id was not valid.


=head1 RESERVATION FUNCTIONS

=item getReservations()

returns a list of all reservations belonging to the specified person_id.

=item getMonthReservations()

finds all reservations for a given person_id in a given month and year.
returns a list of hash-refs of the form: 

  {
	res    => reservation-row from DB,
	bounds => {
		start => start-day,
		end   => end-day,
	}
  }
	
=item getReservation()

given a reservation-id it returns a reference to a reservation-row from
the DB, otherwise undef.

=item addReservation()

adds a new reservation, or modifies an existing one. Parameters
are C<($res_id, $person_id, $start, $end, $type, $desc)>. If
B<$res_id> equals 'new', a new reservation is created, otherwise
B<$res_id> must contain valid reservation-id.

B<$start> and B<$end> must contain dates in the form "D.M.Y".

Returns a list of two items:

=over 4

=item ('disappeared', '')

someone else (presumably) deleted the absence in the meantime

=item ('badday', '')

the start or end day is invalid (i.e., February 30)

=item ('impossible', '')

the start-day is greater than the end-day.

=item ('conflict', $res_id)

there is a conflict with another reservation. B<$res_id> is the
reservation-id of the conflicting reservation.

=item ('ok', '')

success.

=back

=item deleteReservation()

deletes a reservation.  Takes as single parameter a reservation-id.

returns:

=over 4

=item internalerror

The specified reservation-id does not exist.

=item conflict

The reservation-id disappeared suddenly.  Presumably someone else
got there before you.

=item ok

success.

=back


=head1 GROUP FUNCTIONS

=item addGroup()

creates a new group or modifies an existing one.  parameters are
($group_id, $name, $password).  If $group_id equals 'new', a new
group is created, otherwise $group_id must contain a valid group-id.
If an attempt is made to create a group with the same name as an
existing one, returns 'duplicate'.  Returns 'ok' on success.

=item deleteGroup()

guess what this does?  Takes as single parameter the group-id.  Returns
'ok' on success.

=item getGroups()

returns a list of group-ids corresponding to all existing groups.

=item getGroup()

retrieves information about a group.  Takes as single parameter
the group-id.  Return-value depends on context: in a scalar
context, returns group-name, in a list context returns 
a list consisting of B<($group_name, $group_password)>.

=item getGroup()

returns the group-id to which a person belongs.  Takes as single param
the person-id.

=item groupCount()

returns the number of defined groups.

=item getModificationTime()

retrieves modification time for a particular group for a particular
month, if one is known.  This allows C<absence.pl> to determine
if a month-image needs to be created anew.  Returns the time in
epoch-second format if available, otherwise undef.

=back

=head1 AUTHOR

Robert Urban <urban@tru64.org>

Copyright (C) 2003 Robert Urban

=cut
