/*************************************************************************
 ** SQL script to create all tables, views and functions used by
 ** absence.
 **
 ** $Id: create_tables.sql 93 2009-07-14 23:40:07Z urban $
 *************************************************************************/

/********************
 ** GROUP
 ********************/
/*
drop table c_group;
drop sequence c_group_id_seq;
*/
CREATE TABLE c_group
(
	id				SERIAL NOT NULL PRIMARY KEY,
	name			VARCHAR(64) NOT NULL,
	password		VARCHAR(64),
	description		VARCHAR(256),
	invalidated		TIMESTAMP WITH TIME ZONE,
	UNIQUE ( name )
);

CREATE VIEW v_group AS SELECT * FROM c_group WHERE invalidated IS NULL;

/********************
 ** COUNTRY
 ********************/
CREATE TABLE c_country
(
	id				SERIAL NOT NULL PRIMARY KEY,
	code			VARCHAR(2),		/* ISO 2-letter country-code */
	name			VARCHAR(64)		/* long name */
);

/********************
 ** REGION
 ** a region is a part of a country, where has legal holidays that
 ** are valid only for that region
 ********************/
CREATE TABLE c_region
(
	id				SERIAL NOT NULL PRIMARY KEY,
	name			VARCHAR(64),			/* name of region */
	invalidated		TIMESTAMP WITH TIME ZONE,
	description		VARCHAR(256)
);

CREATE VIEW v_region AS SELECT * FROM c_region WHERE invalidated IS NULL;

/********************
 ** OBJECT (usually people)
 ********************/
/*
drop table object;
drop sequence object_id_seq;
*/
CREATE TABLE c_object
(
	id				SERIAL NOT NULL PRIMARY KEY,
	name			VARCHAR(64),
	email			VARCHAR(64),
	description		VARCHAR(256),
	invalidated		TIMESTAMP WITH TIME ZONE,
	country_id		INTEGER REFERENCES c_country(id),
	region_id		INTEGER REFERENCES c_region(id),
	UNIQUE ( name )
);

CREATE VIEW v_object AS SELECT * FROM c_object WHERE invalidated IS NULL;

/********************
 ** GROUP_OBJECT
 ** links objects (people) to groups
 ********************/
CREATE TABLE c_group_object
(
	id				SERIAL NOT NULL PRIMARY KEY,
	group_id		INTEGER NOT NULL REFERENCES c_group(id),
	object_id		INTEGER NOT NULL REFERENCES c_object(id),
	UNIQUE ( group_id, object_id )
);

CREATE INDEX c_group_object_group_id_idx on c_group_object(group_id);
CREATE INDEX c_group_object_object_id_idx on c_group_object(object_id);

/********************
 ** USER
 ********************/
/*
drop table c_user;
drop sequence c_user_id_seq;
*/
CREATE TABLE c_user
(
	id				SERIAL NOT NULL PRIMARY KEY,
 	object_id		INTEGER REFERENCES c_object(id),
	username		VARCHAR(32),
	password		VARCHAR(64),
	pacl			VARCHAR(256),
	gacl			VARCHAR(256),
	invalidated		TIMESTAMP WITH TIME ZONE,
	UNIQUE ( username )
);

CREATE VIEW v_user AS SELECT * FROM c_user WHERE invalidated IS NULL;

CREATE INDEX c_user_username_idx on c_user(username);

/********************
 ** ACCESS
 **
 ** provides access-control when using authorization
 ** 'level' is an integer with following permitted values
 **		1 - read
 **		2 - write
 **		4 - admin
 **
 ** additionally, a group_id of zero implies *all* groups
 ** %% UNUSED CURRENTLY! %%
 ********************/
CREATE TABLE c_access
(
	id				SERIAL NOT NULL PRIMARY KEY,
	user_id			INTEGER NOT NULL REFERENCES c_user(id),
	ref_id			INTEGER,
	type			VARCHAR(10) NOT NULL
		CHECK (type = 'object' OR type = 'group'),
	magic			VARCHAR(10)
		CHECK (magic = 'self' OR magic = 'all'),
	level			INTEGER NOT NULL
);

CREATE INDEX c_access_user_id_idx on c_access(user_id);

/********************
 ** RESERVATION_TYPE
 ********************/
/*
drop table c_reservation_type;
drop sequence c_reservation_type_id_seq;
*/
CREATE TABLE c_reservation_type
(
	id				SERIAL NOT NULL PRIMARY KEY,
	name			VARCHAR(64) NOT NULL,	/* label to be displayed	*/
	height			VARCHAR(16)			/* for overlapping reservations	*/
		CHECK (height IN ('block', 'full', 'half', 'quarter')),
	priority		INTEGER,			/* position within person-bar	*/
	overlap_group	INTEGER,			/* determines which can overlap	*/
	color_red		INTEGER,			/* 0-255 */
	color_green		INTEGER,			/* 0-255 */
	color_blue		INTEGER,			/* 0-255 */
	transparency	INTEGER,			/* 0-127 */
	default_type	BOOLEAN,			/* only one can be marked default */
	skip_non_workdays	BOOLEAN,		/* do not draw reservations on we/hol */
	description		VARCHAR(256),
	invalidated		TIMESTAMP WITH TIME ZONE,
	UNIQUE ( name )
);

CREATE VIEW v_reservation_type AS SELECT * FROM c_reservation_type WHERE invalidated IS NULL;

/********************
 ** NO_OVERLAP
 ** designates which types CANNOT overlap
 ********************/

CREATE TABLE c_no_coincidence
(
	id				SERIAL NOT NULL PRIMARY KEY,
	type_id1		INTEGER NOT NULL REFERENCES c_reservation_type(id),
	type_id2		INTEGER NOT NULL REFERENCES c_reservation_type(id)
);

/********************
 ** RESERVATION
 ********************/
/*
drop table d_reservation;
drop sequence d_reservation_id_seq;
*/
CREATE TABLE d_reservation
(
	id				SERIAL NOT NULL PRIMARY KEY,
 	object_id		INTEGER NOT NULL REFERENCES c_object(id),
 	type_id			INTEGER NOT NULL REFERENCES c_reservation_type(id),
	start			DATE,
	finish			DATE,
	description		VARCHAR(256),
	misc			VARCHAR(256),
	invalidated		TIMESTAMP WITH TIME ZONE
);

CREATE VIEW v_reservation AS SELECT * FROM d_reservation WHERE invalidated IS NULL;

CREATE INDEX d_reservation_start_idx on d_reservation(start);
CREATE INDEX d_reservation_finish on d_reservation(finish);
CREATE INDEX d_reservation_object_id on d_reservation(object_id);

/********************
 ** GROUP_MOD_TIME
 ** epoch-time when group was last modified for specified month/year
 ** i.e., ist group-image still up to date?
 ********************/
CREATE TABLE d_group_mod_time
(
	id				SERIAL NOT NULL PRIMARY KEY,
 	group_id		INTEGER NOT NULL REFERENCES c_group(id),
	month			INTEGER,
	year			INTEGER,
	mod_time		INTEGER,
	image_height	INTEGER,
	image_width		INTEGER,
	UNIQUE ( group_id, month, year )
);
CREATE INDEX d_group_mod_time_group_id_idx on d_group_mod_time(group_id);
CREATE INDEX d_group_mod_time_month_idx on d_group_mod_time(month);
CREATE INDEX d_group_mod_time_year_idx on d_group_mod_time(year);

/********************
 ** CONFIG
 ********************/
CREATE TABLE c_config
(
	id				SERIAL NOT NULL PRIMARY KEY,
	name			VARCHAR(64),
	content			VARCHAR(128)
	/*
	value_type		INTEGER,
	int_value		INTEGER,
	str_value		VARCHAR(256),
	*/
);

CREATE INDEX c_config_name_idx on c_config(name);

/********************
 ** HOLIDAY
 ********************/
CREATE TABLE c_holiday
(
	id				SERIAL NOT NULL PRIMARY KEY,
	country_id		INTEGER REFERENCES c_country(id),
	region_id		INTEGER REFERENCES c_region(id),
	day				DATE,
	description		VARCHAR(256)
);

CREATE INDEX c_holiday_day_idx on c_holiday(day);

/********************
 ** SESSION
 ********************/
CREATE TABLE d_session
(
	id				SERIAL NOT NULL PRIMARY KEY,
	user_id			INTEGER NOT NULL REFERENCES c_user(id),
	ip_addr			INET,
	t_stamp			TIMESTAMP WITH TIME ZONE
);

/********************
 ** RESERVATION-REGION
 ********************/
CREATE TABLE d_reservation_region
(
	id				SERIAL NOT NULL PRIMARY KEY,
 	reservation_id	INTEGER NOT NULL REFERENCES d_reservation(id),
	priority		INTEGER,
	x1				INTEGER,
	y1				INTEGER,
	x2				INTEGER,
	y2				INTEGER
);
CREATE INDEX d_reservation_region_x1_idx on d_reservation_region(x1);
CREATE INDEX d_reservation_region_y1_idx on d_reservation_region(y1);
CREATE INDEX d_reservation_region_x2_idx on d_reservation_region(x2);
CREATE INDEX d_reservation_region_y2_idx on d_reservation_region(y2);


/********************
 ** SESSIONS (for CGI::Session)
 ********************/
CREATE TABLE sessions (
	id				CHAR(32) NOT NULL PRIMARY KEY,
	a_session		BYTEA NOT NULL
);

/*=====================================================================
  = functions
  =====================================================================*/

/* convert a unix epoch-time to postgres timestamp datatype */
create function int2tsz(INTEGER) returns TIMESTAMP WITH TIME ZONE as '
SELECT ( TIMESTAMP WITH TIME ZONE ''epoch'' + $1 * INTERVAL ''1
second'')::TIMESTAMP WITH TIME ZONE;
' language sql;

/* convert a postgres timestamp datatype to unix epoch-time */
create function tsz2int(TIMESTAMP WITH TIME ZONE) returns int as '
select extract( ''epoch'' from $1)::INTEGER;
' language sql;
