package AbsenceImport;

use DBI;
use FileHandle;
use Encode;

use AbsenceUtils;

my $DBH;

END {
	defined($DBH) && $DBH->disconnect;
}

sub init
{
	my ($db_user, $db_pass, $db_name, $db_host) = @_;

	my $cs = "dbi:Pg:dbname=$db_name";
	if (defined($db_host)) {
		$cs .= ";host=$db_host";
	}
	$DBH = DBI->connect($cs,
		$db_user, $db_pass, {AutoCommit => 1, RaiseError => 1});

	defined($DBH) || die "cannot connect to DB";
}

#------------------------------------------------------------------
# loadCountryTable()
# input:
#	expects a reference to array of the form:
#	[
#		{ code => 'CC', name => 'name of country' },
#		{ code => 'C2', name => 'name of country2' },
#		{ code => 'C3', name => 'name of country3' },
#	]
# country names should be encoded as UTF8
#------------------------------------------------------------------
sub loadCountryTable
{
	my $aref = shift;

	my $sql = qq{
		INSERT INTO c_country
		(code,name)
		VALUES (?,?);
	};

	my $sth = $DBH->prepare($sql);
	defined($sth) || die "prepare of [$sql] failed";

	my $rv;
	foreach my $country (@{ $aref }) {
		$rv = $sth->execute($country->{code}, $country->{name});
		defined($rv) || print "error occured for code=$country->{code}, name=$country->{name}\n";
	}

	$sth->finish;
}

sub loadHoliday
{
	my $href = shift;

	my (@fields, @values, @placeholders);

	foreach my $field (keys(%{ $href })) {
		push(@fields, $field);
		push(@values, $href->{$field});
		push(@placeholders, '?');
	}
	my $fields = join(',', @fields);
	my $ph = join(',', @placeholders);

	my $sql = qq{
		INSERT INTO c_holiday
		($fields)
		VALUES ($ph);
	};
	#print "SQL: $sql";

	$DBH->do($sql, undef, @values);

	my $id = getCurrval('c_holiday_id_seq');
	return $id;
}

sub lookupCountry
{
	my $code = uc(shift);

	my $sql = qq{
		SELECT id
		FROM c_country
		WHERE UPPER(code) = '$code';
	};

	my $ref = dbSelect($sql);
	return defined($ref) ? $ref->{id} : undef;
}

sub lookupRegion
{
	my $name = uc(shift);

	my $sql = qq{
		SELECT id
		FROM c_region
		WHERE UPPER(name) = '$name';
	};

	my $ref = dbSelect($sql);
	return defined($ref) ? $ref->{id} : undef;
}

sub createRegion
{
	my ($name, $desc) = @_;

	my $fields = 'name';
	my $values = "'$name'";
	if (defined($desc)) {
		$fields .= ',description';
		$values .= ",'$desc'";
	}

	my $sql = qq{
		INSERT INTO c_region
			($fields)
		VALUES
			($values);
	};

	#$DBH->execute($sql);

	my $sth = $DBH->prepare($sql);
	$sth->execute || die "execute on sql failed:\n$sql\n--end--\n";
	$sth->finish;

	my $id = getCurrval('c_region_id_seq');
	return $id;
}

sub getCurrval
{
	my $sequence = shift;

	my $sql = qq{SELECT currval('$sequence');};
	my $sth = $DBH->prepare($sql);
	$sth->execute || die "failed to get currval for $sequence\n";
	my $id = ($sth->fetchrow_array)[0];
	defined($id) || die "no ID returned.";
	return $id;
}

sub getDbHolidays
{
	my $sql = qq{
		SELECT * FROM c_holiday ORDER BY day;
	};

	return dbSelect($sql);
}

my %COUNTRY_CACHE;
sub getCountry
{
	my $id = shift;

	if (exists($COUNTRY_CACHE{$id})) {
		return $COUNTRY_CACHE{$id};
	}

	my $sql = qq{
		SELECT * FROM c_country WHERE id = $id;
	};

	return dbSelect($sql);
}

sub getCountries
{
	my $sql = qq{
		SELECT * FROM c_country;
	};

	return dbSelect($sql);
}

my %REGION_CACHE;
sub getRegion
{
	my $id = shift;

	if (exists($REGION_CACHE{$id})) {
		return $REGION_CACHE{$id};
	}

	my $sql = qq{
		SELECT * FROM c_region WHERE id = $id;
	};

	return dbSelect($sql);
}

sub getRegions
{
	my $sql = qq{
		SELECT * FROM c_region WHERE invalidated IS NULL;
	};

	return dbSelect($sql);
}

sub lookupHoliday
{
	my ($d, $m, $y, $addit) = @_;

	my %qual = (
		country_id	=> undef,
		region_id	=> undef,
	);
	if (defined($addit)) {
		if (!exists($qual{ $addit->[0] })) {
			die "qualifier [$addit->[0]] unknown";
		}
		$qual{ $addit->[0] } = $addit->[1];
	}

	my $day = sprintf('%04d-%02d-%02d', $y, $m, $d);
	
	my $sql = qq{
		SELECT id FROM c_holiday WHERE day = '$day'
	};
	foreach my $q (keys(%qual)) {
		if (defined($qual{$q})) {
			$sql .= " AND $q = $qual{$q}";
		}
		else {
			$sql .= " AND $q IS NULL";
		}
	}
	$sql .= ';';

	my $ref = dbSelect($sql);
	return defined($ref) ? $ref->{id} : undef;
}

sub deleteHoliday
{
	my $id = shift;

	my $sql = qq{
		DELETE FROM c_holiday WHERE id = $id;
	};

	my $res = $DBH->do($sql);
	defined($res) || die "delete of id [$id] from c_holiday failed\n";
}

sub deleteAllHolidays
{
	my $id = shift;

	my $sql = qq{
		DELETE FROM c_holiday;
	};

	my $res = $DBH->do($sql);
	defined($res) || die "delete of id [$id] from c_holiday failed\n";

	$sql = qq{ ALTER SEQUENCE c_holiday_id_seq RESTART WITH 1; };
	$res = $DBH->do($sql);
	defined($res) || die "restart sequence c_holiday_id_seq failed";
}

sub deleteRegion
{
	my $id = shift;

	my $now = time();

	my $sql = qq{
		UPDATE c_region SET invalidated = int2tsz($now) WHERE id = $id;
	};

	my $res = $DBH->do($sql);
	defined($res) || die "delete of id [$id] from c_region failed\n";
}

sub parseIcal
{
	my $file = shift;

	my $fh = FileHandle->new($file);
	defined($fh) || die "failed to open file [$file] for reading: $!";

	my $curr_event;
	my $state;
	my $data;
	my $debug = 0;

	while(<$fh>) {
		s/[\n\r]+$//;
		$debug && print "[$_]\n";
		if (!defined($state)) {
			if (!/^BEGIN:VCALENDAR$/) {
				die "file does not begin with 'BEGIN:VCALENDAR'\n";
			}
			$state = 'calendar';
			$debug && print "  state = calendar\n";
		}
		elsif ($state eq 'calendar') {
			if (/^BEGIN:VEVENT$/) {
				$state = 'event';
				$debug && print "  state = event\n";
				$curr_event = {};
				next;
			}
		}
		elsif ($state eq 'event') {
			if (/^END:VEVENT/) {
				$state = 'calendar';
				$debug && print "  state = calendar\n";
				exists($curr_event->{date}) || next;
				my $d = $curr_event->{date};
				if (exists($curr_event->{fixed})) {
					$d->{year} = '*';
				}
				$data->{ $d->{year} }->{ $d->{month} }->{ $d->{day} }
					= $curr_event->{summary};
				next;
			}
			if (/^SUMMARY:(.*)$/) {
				$curr_event->{summary} = decode_utf8($1);
				$debug && print "  summary = [$1]\n";
			}
			elsif (/^DESCRIPTION:(.*)$/) {
				$curr_event->{description} = decode_utf8($1);
			}
			elsif (/^DTSTART(;[^:]+)?:(.*)$/) {
				if ($1 eq ';VALUE=DATE') {
					my ($y, $m, $d) = ($2 =~ /^(\d{4})(\d{2})(\d{2})$/);
					$curr_event->{date} = { year => $y, month => $m, day => $d };
					$debug && print "  date = [$y/$m/$d]\n";
				}
			}
			elsif (/^RRULE:/ && /FREQ=YEARLY/i) {
				$curr_event->{fixed} = 1;
			}
		}
	}
	$fh->close;

	return $data;
}

1;
###########################
sub dbSelect
{
	my ($sql, @params) = @_;

	my $sth;
	eval {
		$sth = $DBH->prepare($sql);
	};
	if ($@) {
		die $@;
	}

	$sth->execute(@params) || die "statement [$sql] failed";

	my @result;
	my $ref;
	while(defined($ref = $sth->fetchrow_hashref)) {
		push(@result, $ref);
	}

	if (wantarray) {
		return @result;
	} else {
		@result || return undef;
		return $result[0];
	}
}

1;
