package AbsenceMigration;

use FileHandle;
use Encode qw(encode decode from_to decode_utf8);

use AbsenceUtils;
#use AbsenceInstall;

use strict;

my $DEBUG = 0;
my $DBH;
my $LOGFH;

my %PERSON_KEYS = (
	name		=> '',
	id			=> '',
	user_id		=> '',
	email		=> '',
	group		=> '',		# reference to list of GIDs
);
my @RES_KEYS		= qw(id person_id start end type desc misc);
my %RESERVATION_KEYS = (
	id			=> '',		# unique number
	person_id	=> '',		# unique number
	start		=> '',		# date. array-ref in form [d,m,y]
	end			=> '',		# date. array-ref in form [d,m,y]
	type		=> '',		# single word "vacation", "busy", "travel"
	desc		=> '',		# arbitrary text
	misc		=> '',		# more arbitrary text
);
my %GROUP_KEYS = (
	id			=> '',
	name		=> '',
	pass		=> '',
);
my %USER_KEYS = (
	user_id		=> '',
	password	=> '',
	pacl		=> '',
	gacl		=> '',
);

sub migrateData
{
	my $config_ref = shift;

	my $logfile = "/tmp/absence-migration-$$.log";
	openlog($logfile);
	my $data_dir = getDataDir();
	miglog("data-dir = [$data_dir]", 1);

	miglog("** migrating types **", 1);
	my $type_map = migrateTypes($data_dir);

	my $db_file = "$data_dir/absence.db";
	-f $db_file || die "db-file not found";

	miglog("** reading data file **", 1);
	my $data = readDbFile($db_file);

	# migrate groups
	miglog("** migrating groups **", 1);
	my $group_info = migrateGroups($data->{groups});

	# migrate people
	miglog("** migrating people **", 1);
	migratePeople($config_ref, $data, $group_info);
	if ($config_ref->{authentication} eq 'yes') {
		miglog("** migrating users **", 1);
		migrateUsers($config_ref, $data);
		miglog("** migrating ACLs **", 1);
		migrateAcls($config_ref, $data, $group_info);
	}

	# migrate reservations!
	miglog("** migrating reservations **", 1);
	migrateReservations($config_ref, $data, $type_map);

	closelog();
	print qq{log written to "$logfile".\n};
}

sub migrateReservations
{
	my ($config_ref, $data, $type_map) = @_;

	#my $dbh = connectDb(
	#	$config_ref->{database_user},
	#	$config_ref->{database_pass},
	#	$config_ref->{database_name},
	#	$config_ref->{database_host},
	#	undef,
	#);
	#defined($dbh) || die "failed to connect to DB";

	my $migrated = 0;

	foreach my $rid (keys(%{ $data->{reservations} })) {
		miglog("  rid=$rid");
		my $rref = $data->{reservations}->{$rid};
		my $start = {
			day		=> $rref->{start}->[0],
			month	=> $rref->{start}->[1],
			year	=> $rref->{start}->[2],
		};
		my $end = {
			day		=> $rref->{end}->[0],
			month	=> $rref->{end}->[1],
			year	=> $rref->{end}->[2],
		};
		my $type = $rref->{type};
		if (!exists($type_map->{$type})) {
			miglog("    cannot migrate res=$rid, type [$type] not found");
			next;
		}
		my $type_id = $type_map->{$type};
		my $pid = $rref->{person_id};
		if (!exists($data->{people}->{$pid})) {
			miglog("    cannot migrate res=$rid, pid [$pid] not found");
			next;
		}
		if (!exists($data->{people}->{$pid}->{new_pid})) {
			miglog("    cannot migrate res=$rid, pid [$pid] was not migrated");
			next;
		}
		my $new_pid = $data->{people}->{$pid}->{new_pid};
		my $res;
		my $desc = convertString($rref->{desc});
		eval {
			$res = AbsenceDB::_addReservation(
				'new', $new_pid, $start, $end, $type_id, $desc
			);
		};
		if ($@) {
			miglog("    failed to migrate res=$rid.");
		}
		elsif (defined($res)) {
			miglog("    successfully migrated reservation.");
			$migrated++;
		}
	}
	miglog("    migrated $migrated reservations", 1);
}

sub _convertString
{
	my $str = shift;

	my $copy = $str;
	my $decoded;
	eval {
		$decoded = decode('utf8', $copy, 1);
	};
	if ($@) {
		# decode failed. assume iso-8859-1
		$decoded = decode('iso-8859-1', $str);
	}

	return $decoded;
	#return encode('utf8', $decoded);
}

sub convertString
{
	my $str = shift;

	return encode('utf8', $str);
}

sub convertDate
{
	my $ref = shift;

	return join('-', reverse(@{ $ref }));
}

sub migrateTypes
{
	my $data_dir = shift;

	#--------------------------------------------------------
	# types first
	#--------------------------------------------------------
	my $types_file = "$data_dir/absence-types.db";
	-f $types_file || die "types-file not found";

	miglog("  opening types file");
	my $types = readTypesFile($types_file);

	my $migrated = 0;
	my $map;
	my $ref;
	foreach my $type_name (keys(%{ $types->{db} })) {
		miglog("  migrating type [$type_name]");
		my @cols = @{ $types->{db}->{$type_name} };
		$ref = {
			name		=> convertString($type_name),
			color_red	=> $cols[0],
			color_green	=> $cols[1],
			color_blue	=> $cols[2],
			height		=> 'full',
			priority	=> 1,
		};
		if (defined($types->{default}) && ($types->{default} eq $type_name)) {
			$ref->{default_type} = 't';
		}
		my $id = AbsenceDB::addType($ref);
		miglog("    type created with id=$id");
		$migrated++;
		$map->{$type_name} = $id;
	}

	miglog("    migrated $migrated types.", 1);
	return $map;
}

sub migrateGroups
{
	my $groups = shift;

	my $group_info;

	my $migrated = 0;

	foreach my $gid (keys(%{ $groups })) {
		my $name = convertString($groups->{$gid}->{name});
		miglog("  migrating group [$name] with gid=$gid");
		my ($res, $new_gid) = AbsenceDB::addGroup('new', $name, undef, undef);
		if ($res eq 'ok') {
			$group_info->{$gid} = { name => $name, new_id => $new_gid };
			miglog("    new gid = $new_gid");
			$migrated++;
		}
		else {
			miglog("    failed to migrate group. res=[$res]");
		}
	}

	miglog("    migrated $migrated groups.", 1);
	return $group_info;
}

sub migratePeople
{
	my ($config_ref, $data, $group_info) = @_;

	my $auth = $config_ref->{authentication} eq 'yes';

	my @params;

	my $migrated = 0;

	foreach my $pid (keys(%{ $data->{people} })) {
		my $pref = $data->{people}->{$pid};
		miglog("  processing pid=$pid ($pref->{name})");
		my $gref = convertGroups($group_info, $pref->{group});
		miglog("    groups: old=".join(',', @{ $pref->{group} })
			." new=".join(',', @{ $gref }));
		my $uref;
		@params = (
			'new',
			convertString($pref->{name}),
			'',
			$pref->{email},
			$gref,
		);
		if ($auth
			&& exists($pref->{user_id})
			&& exists($data->{users}->{ $pref->{user_id} }))
		{
			my $user = $pref->{user_id};
			$uref = $data->{users}->{$user};
			$uref->{attached} = 1;
			push(@params, $uref->{password}, undef, undef);
			$params[2] = $user;
		}
		elsif ($auth) {
			if (!exists($pref->{user_id})) {
				miglog("    no user in DB for person [$pref->{name}]");
			} elsif (!exists($data->{users}->{ $pref->{user_id} })) {
				miglog("    cannot find user [$pref->{user_id}] in USERS");
			}
		}
		my ($res, $new_pid, $uid) = AbsenceDB::addPerson(@params);
		if ($res eq 'ok') {
			miglog("    person [$pref->{name}] migrated. new pid = $new_pid");
			$pref->{new_pid} = $new_pid;
			$uref->{new_uid} = $uid;
			$migrated++;
		}
		else {
			miglog("    failed to create person [$pref->{name}]");
		}
	}
	miglog("    migrated $migrated people.", 1);
}

sub migrateUsers
{
	my ($config_ref, $data) = @_;

	my $migrated = 0;

	foreach my $username (keys(%{ $data->{users} })) {
		my $uref = $data->{users}->{$username};
		miglog("  processing user [$username]");
		exists($uref->{attached}) && next;
		$username = convertString($username);
		my ($res, $uid) = AbsenceDB::addUser($username, $uref->{password}, undef, undef);
		if ($res eq 'ok') {
			miglog("  user [$username] migrated. new uid = $uid");
			$uref->{new_uid} = $uid;
			$migrated++;
		}
		else {
			miglog("    failed to migrate user [$username]. res=[$res]");
		}
	}
	miglog("    migrated $migrated people.", 1);
}

sub migrateAcls
{
	my ($config_ref, $data, $group_info) = @_;

	foreach my $username (keys(%{ $data->{users} })) {
		my $uref = $data->{users}->{$username};
		miglog("  processing ACLs for user [$username]");
		if (!exists($uref->{new_uid})) {
			miglog("    skipping acls for user [$username] because migrate failed");
			next;
		}
		my ($pacl_ref, $gacl_ref)
			= convertAcls($group_info, $data, $uref->{pacl}, $uref->{gacl});
		miglog("  migrating ACLs for user [$username]");
		AbsenceDB::setAcls($uref->{new_uid}, $pacl_ref, $gacl_ref);
	}
}

#--------------------------------------------------------------------
# convertAcls()
#
# gets old pacls and gacls in the form "[war]:(self|all|ID)"
# and converts them to form
# $ref = {
#	target	=> (self|all|ID),
#	level	=> 1|2|4,
# }
#--------------------------------------------------------------------
sub convertAcls
{
	my ($group_info, $data, $old_pacl, $old_gacl) = @_;

	my %levmap = (r => 1, w => 2, a => 4);

	my ($paclr, $gaclr);

	my @new_pacl;
	foreach my $acl (@{ $old_pacl }) {
		my ($perm, $pid) = ($acl =~ /^([war]):(.*)$/);
		if (!defined($perm)) {
			miglog("    failed to match ACL [$acl]");
			next;
		}
		my $level = $levmap{$perm};
		if ($pid =~ /^(all|self)$/) {
			push(@new_pacl, {
				target	=> $pid,
				level	=> $level,
			});
			next;
		}
		if (!exists($data->{people}->{$pid})) {
			miglog("    pid [$pid] unknown.");
			next;
		}
		my $pref = $data->{people}->{$pid};
		if (!exists($pref->{new_pid})) {
			miglog("    skipping ACL [$acl] because migration failed.");
		}
		push(@new_pacl, {
			target	=> $pref->{new_pid},
			level	=> $level,
		});
	}

	my @new_gacl;
	foreach my $acl (@{ $old_gacl }) {
		my ($perm, $gid) = ($acl =~ /^([war]):(.*)$/);
		if (!defined($perm)) {
			miglog("    failed to match ACL [$acl]");
			next;
		}
		my $level = $levmap{$perm};
		if ($gid =~ /^(all|self)$/) {
			push(@new_gacl, {
				target	=> $gid,
				level	=> $level,
			});
			next;
		}
		if (!exists($group_info->{$gid})) {
			miglog("    gid [$gid] unknown.");
			next;
		}
		if (!exists($group_info->{$gid})) {
			miglog("    skipping ACL [$acl] because no group mapping.");
			next;
		}
		my $gref = $group_info->{$gid};

		push(@new_gacl, {
			target	=> $gref->{new_id},
			level	=> $level,
		});
	}

	return (\@new_pacl, \@new_gacl);
}

sub convertGroups
{
	my ($group_info, $gref) = @_;

	my @result;
	foreach my $old (@{ $gref }) {
		next unless (exists($group_info->{$old}));
		push(@result, $group_info->{$old}->{new_id});
	}

	return \@result;
}

sub getDataDir
{
	my $data_dir;
	my $finished = 0;
	do {
		$data_dir = query(
			'Directory where old datafiles are located',
			'',
			undef,
			0,
			0,
		);
		if (! -d $data_dir) {
			print "[$data_dir] is not a directory, or does not exist\n";
		} elsif (! -f "$data_dir/absence.db") {
			print qq{file "absence.db" not found in directory [$data_dir]\n};
		} else {
			$finished = 1;
		}
	} while(!$finished);

	return $data_dir;
}

sub readHolidays
{
	my $file = shift;

	my ($day, $month, $year, $desc);
	my $data;

	my $fh = FileHandle->new($file);
	defined($fh) || die "open [$file] for reading";
	while(<$fh>) {
		chomp;
		($DEBUG > 1) && print "HOLIDAY> $_\n";
		if (/^\s*$/) { next; }
		if (/^\s*#$/) { next; }
		if (/^(\d+)\.(\d+).(\d+|\*)\t(\S+(\s+\S+)*)\s*$/) {
			($day, $month, $year, $desc) = ($1, $2, $3, $4);
			$DEBUG && print "match: d=[$day], m=[$month], y=[$year], d=[$desc]\n";
			#push(@{$DB{$month}->{$year}}, [$day, $month, $year, $desc]);
			$data->{$year}->{$month}->{$day} = _convertString($desc);
		}
	}
	$fh->close;

	return $data;
}

sub readDbFile
{
	my $file = shift;

	my $mode = '';
	my $in_block = 0;
	my ($key, $val);
	my $curr;

	$DEBUG && miglog("read file.");
	my $fh = FileHandle->new($file);
	defined($fh) || die "readDbFile: failed to open file [$file]: $!";

	$DEBUG && miglog("reading DB");

	my %GROUPS			= ();
	my %PEOPLE			= ();
	my %RESERVATIONS	= ();
	my %USERS			= ();

	while(<$fh>) {
		chomp;
		$_ = decode('iso-8859-1', $_);
		($DEBUG > 2) && print "LINE: $_\n";
		if (/^\s*#/) { next; }		# skip WS
		if (/^PEOPLE:/) {
			$mode = 'people';
			next;
		} elsif (/^RESERVATIONS:/) {
			$mode = 'reservations';
			($DEBUG > 2) && print "found RESERVATIONS:\n";
			next;
		} elsif (/^GROUPS:/) {
			$mode = 'groups';
			($DEBUG > 2) && print "found GROUPS:\n";
			next;
		} elsif (/^USERS:/) {
			$mode = 'users';
			($DEBUG > 2) && print "found USERS:\n";
			next;
		}
	
		if ($mode eq 'people') {
			if ($in_block) {
				if (/^END/) {
					$in_block = 0;
					if (!exists($curr->{id})) {
						die "no id found for person [$curr->{name}]";
					}
					if (exists($PEOPLE{$curr->{id}})) {
						die "person ID [$curr->{id}] occurs twice";
					}
					if (!exists($curr->{group})) {
						$curr->{group} = [1];
					}
					$PEOPLE{$curr->{id}} = $curr;
					undef $curr;
					next;
				}
				if (/^\s*(\S+):\s*(\S+(\s+\S+)*)\s*$/) {
					($key, $val) = ($1, $2);
					if (exists($PERSON_KEYS{$key})) {
						if ($key eq 'group') {
							$curr->{$key} = [split(/,/, $val)];
						} else {
							$curr->{$key} = $val;
						}
					}
				}
			} elsif (/^START/) {
				$in_block = 1;
			}
		} elsif ($mode eq 'reservations') {
			if ($in_block) {
				if (/^END/) {
					$in_block = 0;
					($DEBUG > 2) && print "END: id = $curr->{id}\n";
					if (!exists($curr->{id})) {
						die "no id found for reservation [$curr->{start}]";
					}
					if (exists($RESERVATIONS{$curr->{id}})) {
						die "reservation ID [$curr->{id}] occurs twice";
					}
					$curr->{sjd} = AbsenceDate::julianDay(@{$curr->{start}});
					$curr->{ejd} = AbsenceDate::julianDay(@{$curr->{end}});
					$RESERVATIONS{$curr->{id}} = $curr;
					undef $curr;
					next;
				}
				if (/^\s*(\S+):\s*(\S+(\s+\S+)*)\s*$/) {
					($key, $val) = ($1, $2);
					($DEBUG > 2) && print "key = [$key], val = [$val]\n";
					if (exists($RESERVATION_KEYS{$key})) {
						($DEBUG > 2) && print "setting \$curr->{$key} = [$val]\n";
						if ($key =~ /^(start|end)$/) {
							$curr->{$key} = [split(/\./,$val)];
						} else {
							$curr->{$key} = $val;
						}
					}
				}
			} elsif (/^START/) {
				$in_block = 1;
				next;
			}
		} elsif ($mode eq 'groups') {
			if ($in_block) {
				if (/^END/) {
					$in_block = 0;
					($DEBUG > 2) && print "END: id = $curr->{id}\n";
					if (!exists($curr->{id})) {
						die "no id found for group [$curr->{name}]";
					}
					if (exists($GROUPS{$curr->{id}})) {
						die "group ID [$curr->{id}] occurs twice";
					}
					$GROUPS{$curr->{id}} = $curr;
					#$GROUP_COUNT++;
					undef $curr;
					next;
				}
				if (/^\s*(\S+):\s*(\S+(\s+\S+)*)\s*$/) {
					($key, $val) = ($1, $2);
					($DEBUG > 2) && print "key = [$key], val = [$val]\n";
					if (exists($GROUP_KEYS{$key})) {
						($DEBUG > 2) && print "setting \$curr->{$key} = [$val]\n";
						$curr->{$key} = $val;
					}
				}
			} elsif (/^START/) {
				$in_block = 1;
				next;
			}
		} elsif ($mode eq 'users') {
			if ($in_block) {
				if (/^END/) {
					$in_block = 0;
					($DEBUG > 2) && print "END: id = $curr->{user_id}\n";
					if (!exists($curr->{user_id})) {
						die "no id found for user [$curr->{user_id}]";
					}
					if (exists($USERS{$curr->{user_id}})) {
						die "user ID [$curr->{user_id}] occurs twice";
					}
					$USERS{$curr->{user_id}} = $curr;
					#$USER_COUNT++;
					undef $curr;
					next;
				}
				if (/^\s*(\S+):\s*(\S+(\s+\S+)*)\s*$/) {
					($key, $val) = ($1, $2);
					($DEBUG > 2) && print "key = [$key], val = [$val]\n";
					if ($key =~ /acl$/) {
						$val = [split(/,/, $val)];
					}
					if (exists($USER_KEYS{$key})) {
						($DEBUG > 2) && print "setting \$curr->{$key} = [$val]\n";
						$curr->{$key} = $val;
					}
				}
			} elsif (/^START/) {
				$in_block = 1;
				next;
			}
		}
	}

	$fh->close;

	return {
		groups			=> \%GROUPS,
		people			=> \%PEOPLE,
		reservations	=> \%RESERVATIONS,
		users			=> \%USERS,
	};
}

sub readTypesFile
{
	my $file = shift;

	my $default;
	my %db;

	my $fh = FileHandle->new($file);
	defined($fh) || die "open $file for reading";
	my ($type, $color);
	while(<$fh>) {
		if (/^\s*#/) { next; }
		if (/^\s*$/) { next; }
		if (/^Type:\s+(\S+)\s+(\S+)\s+(\S+)\s+(\S+)\s*(#.*)?$/) {
			$type = $1;
			my @cols = ($2, $3, $4);
			$color = [colcon(@cols)];	# rgb
			$db{$type} = $color;
		} elsif (/^\s*Default: (\S+)\s*$/) {
			$default = $1;
		}
	}
	$fh->close;

	return { db => \%db, default => $default };
}

sub colcon
{
	my (@list) = @_;

	my @new;
	foreach my $col (@list) {
		($col =~ /^0x/i) ? push(@new, hex($col)) : push(@new, $col);
	}

	@new;
}

sub openlog
{
	my $file = shift;

	$LOGFH = FileHandle->new($file, 'w');
	defined($LOGFH) || die "failed to open logfile [$file] for writing: $!";
}

sub closelog
{
	$LOGFH->close;
}

sub miglog
{
	my ($msg, $show) = @_;

	print $LOGFH "$msg\n";
	$show && print "$msg\n";
}

1;
