package AbsenceUtils;

use FileHandle;
use IPC::Open3;
use IO::Select;
use POSIX ":sys_wait_h";
use English;
use DBI;
use Encode;

use Exporter;
@ISA = qw(Exporter);
@EXPORT = qw(query yesNo inList loadAbsenceDB connectDb executeSql runCommand);
@EXPORT_OK = qw(query yesNo inList loadAbsenceDB connectDb executeSql runCommand);

use strict;

# package globals

my $VERBOSE	= 1;
my $DEBUG	= 0;

sub loadAbsenceDB
{
	my ($libdir, $logfile) = @_;

	my $found = 0;
	foreach my $dir (@INC) {
		if ($dir eq $libdir) {
			$found = 1;
		}
	}

	if (!$found) {
		push(@INC, $libdir);
	}

	$logfile = '/dev/null' unless(defined($logfile));

	eval {  
		require "$libdir/AbsenceConfig.pm";
		require "$libdir/AbsenceLog.pm";
		AbsenceLog->import(LOGFILE => $logfile);
		require "$libdir/AbsenceDB.pm";
	};
	if ($@) {
		print "Error while trying to load absence modules:\n$@\n";
		exit 1;
	}
}

sub query
{
	my ($prompt, $default, $allowed, $none, $undef) = @_;

	my $ans;
	my $def = length($default) ? " [$default]" : '';
	if (ref($allowed)) {
		($none || $undef) && push(@{$allowed}, 'none');
		$prompt .= ' ('.join('/', @{$allowed}).')';
	} elsif ($none || $undef) {
		$prompt .= ' ("none" for none)';
	}
	my $qs = "$prompt${def}? ";
	while(!length($ans)) {
		print "$qs";
		chomp($ans = <STDIN>);
		if (!length($ans) && $default) {
			$ans = $default;
		}
		if ($ans) {
			if ($allowed) {
				if (ref($allowed)) {
					if (!inList($ans, $allowed)) {
						print 'must be one of ('.join(',', @{$allowed}).")\n";
						$ans = '';
					}
				} elsif ($none && ($ans eq 'none')) {
					$ans = '';
					last;
				} elsif ($undef && ($ans eq 'undef' || $ans eq 'none')) {
					$ans = undef;
					last;
				} elsif ($ans !~ /^$allowed$/) {
					print "not a valid answer.\n";
					$ans = '';
				}
			}
		}
	}

	return $ans;
}

sub yesNo
{
	my ($prompt, $default) = @_;

	if ($default =~ /^1|0$/) {
		$default = ($default == 1) ? 'yes' : 'no';
	}
	my $ans = query($prompt, $default, ['yes', 'no', 'y', 'n']);
	if ($ans =~ /^y/i) {
		return 1;
	}
	return 0;
}

sub inList
{
	my ($thing, $lref) = @_;

	foreach (@{$lref}) {
		if ($thing eq $_) {
			return 1;
		}
	}

	0;
}

sub connectDb
{
	my ($user, $pass, $name, $host, $port) = @_;

	my $cs = "dbi:Pg:dbname=$name";
	if (defined($host)) {
		$cs .= ";host=$host";
	}

	if (defined($port)) {
		$cs .= ";port=$port";
	}

	my $dbh = DBI->connect($cs, $user, $pass,
		{AutoCommit => 1, RaiseError => 0});
	
	return $dbh;
}

sub executeSql
{
	my ($dbh, $sql) = @_;

	($VERBOSE > 1) && print "SQL to be executed:\n$sql\n";
	$DEBUG && return 1;

	return defined($dbh->do($sql)) ? 1 : 0;
}

#-------------------------------------------------------------------------
# runCommand()
# ------------
# runs an external command, optionally providing input via STDIN. reads
# STDOUT and STDERR separately.
#
# params:
#	-cmd	=> [ 'external-command', 'and', 'its', 'params']
#	-su 	=> [ 'user', 'group' ]	# switch to user/group before running
#	-input	=> [ @lines_of_input ]	# lines will be fed to program (NL is added)
#-------------------------------------------------------------------------
sub runCommand
{
	my %args = @_;

	my $debug = 0;

	my @cmd = @{ $args{-cmd} };

	my $reset_uid = 0;
	my ($old_euid, $old_egid);
	if (exists($args{-su})) {
		($old_euid, $old_egid) = ($EUID, $EGID);
		(ref($args{-su}) eq 'ARRAY') || die "must supply [user, group] for -su";

		my $new_gid = getgrnam($args{-su}->[1]);
		defined($new_gid) || die "group [$args{-su}->[1]] unknown\n";
		$EGID = "$new_gid $new_gid";

		my $new_uid = getpwnam($args{-su}->[0]);
		defined($new_uid) || die "user [$args{-su}->[0]] unknown\n";
		$EUID = $new_uid;

		$reset_uid = 1;
	}

	my @input = exists($args{-input}) ? @{$args{-input}} : ();
	my ($read_fh, $write_fh, $err_fh);

	$debug && print "run_command()\n";

	$debug && print "CMD: [", join(' ', @cmd), "]\n";

	my $child_status;

	$SIG{PIPE} = 'IGNORE';

	$SIG{CHLD} = sub {
		$debug && print "got SIGCHLD!!!\n";
		if (waitpid(-1, WNOHANG) > 0) {
			$child_status = $?;
		}
	};

	$err_fh = FileHandle->new;	# only reader and writer are auto-generated
	$debug && print "calling open3...\n";
	my $pid = open3($write_fh, $read_fh, $err_fh, @cmd);

	# read output until EOF
	my ($rout, $rin, $wout, $win, $eout, $ein);
	$rin = $ein = '';
	my ($buf, $ret, $out, $err);

	my ($fileno_write, $fileno_err, $fileno_read);

	my $sel_r = IO::Select->new;
	my $sel_w = undef;

	if (@input) {
		$sel_w = IO::Select->new($write_fh);
		$win = '';
		$debug && print "have input, leaving input-fh open...\n";
	} else {
		$debug && print "no input, closing input-fh ...\n";
		close($write_fh);
	}

	if (fileno($read_fh)) {
		$sel_r->add($read_fh);
		$debug && print "read-fh defined...\n";
	}

	if (fileno($err_fh)) {
		$sel_r->add($err_fh);
		$debug && print "err-fh defined...\n";
	}

	my $input_line;
	my %result;

	$debug && print "  going into read loop...\n";
	while ($sel_r->handles) {
		$debug && print "\n**** top of while ****\n";

		$debug && print "going into select...\n";
		my @ready = IO::Select->select($sel_r, $sel_w, undef, undef);

		# READ
		if (@{$ready[0]}) {
			for my $fh (@{ $ready[0] }) {
				my $label = ($fh == $read_fh) ? 'out' : 'err';
				$debug && print "std$label has something...\n";
				$ret = sysread($fh, $buf, 512);
				$debug && print "read [$ret] bytes\n";
				if ($ret == 0) {
					$debug && print "hit end of STD$label. closing\n";
					$sel_r->remove($fh);
					$fh->close;
				} else {
					$result{$label} .= $buf;
					$debug && print "[$label] = $buf\n";
				}
			}
		}

		# WRITE
		if (@{$ready[1]}) {
			my $fh = $ready[1][0];
			$debug && print "stdin is ready for input...\n";
			$input_line = shift(@input)."\n";
			$debug && print "INPUT: [$input_line]\n";
			$ret = syswrite($write_fh, $input_line);
			defined($ret) || die "write failed: $!\n";
			$debug && print "wrote [$ret] bytes\n";
			if (!@input) {
				$debug && print "input exhausted. closing\n";
				$sel_w->remove($write_fh);
				$write_fh->close;
			}
		}
	}

	$debug && print "waitpid returned status [$child_status]\n";

	if ($reset_uid) {
		$EUID = $old_euid;
		$EGID = $old_egid;
	}

	return ($child_status/256, $result{out}, $result{err});
}

sub convertString
{
	my $str = shift;

	my $copy = $str;
	my $decoded;
	eval {
		$decoded = decode('utf8', $copy, 1);
	};
	if ($@) {
		# decode failed. assume iso-8859-1
		$decoded = decode('iso-8859-1', $str);
	}

	return encode('utf8', $decoded);
}

1;
